from . import cli
from gallama.logger.logger import logger
from gallama.backend.thinking_template import THINKING_TEMPLATE, Thinking
from . import app
from .app import make_server
from .backend import chatgenerator
from .config import ConfigManager


# using absolute import here as this file will be run alone
import torch
import gc
from fastapi import FastAPI, Request, APIRouter
from fastapi.middleware.cors import CORSMiddleware
from gallama.data_classes.data_class import (
    ChatMLQuery,
    ToolForce,
    GenerateQuery,
    ModelObjectResponse,
    ModelObject,
    ModelParser,
    GenQueue,
    EmbeddingRequest
)
import argparse
from gallama.backend.model import Model
from gallama.backend.prompt_engine import PromptEngine
from gallama.backend.chatgenerator import ChatGenerator
import uvicorn
from fastapi.exceptions import RequestValidationError
from sse_starlette.sse import EventSourceResponse
import json
from fastapi import HTTPException
from fastapi.responses import JSONResponse
from gallama.api_response.chat_response import (
    chat_completion_response_stream,
    completion_response_stream,
    completion_response,
    chat_completion_response,
    chat_completion_response_artifact_stream,
    chat_completion_response_artifact
)
from gallama.config.config_manager import ConfigManager
from logging import DEBUG
from gallama.logger.logger import get_logger
from gallama.data import ARTIFACT_SYSTEM_PROMPT
import os
import asyncio
from contextlib import asynccontextmanager

try:
    from gallama.backend.chatgenerator import ChatGeneratorLlamaCpp
except ImportError:
    # llama cpp optional dependency
    ChatGeneratorLlamaCpp = None

# experimental feature: tensor parallel
try:
    from exllamav2 import ExLlamaV2Cache_TP
except:
    # optional dependency
    ExLlamaV2Cache_TP = None


# Add this after your imports to clear logging from 3rd party module

#streaming example
#https://blog.gopenai.com/how-to-stream-output-in-llm-based-applications-to-boost-user-experience-e9fcf582777a

logger = get_logger()
router = APIRouter()
config_manager = ConfigManager()
model_ready = False


@router.get("/")
async def read_root():
    return {"Hello": "World"}


def validate_api_request(query: ChatMLQuery):
    # fixing api call parameter
    # update temperature to 0.1 because huggingface doesn't accept 0 temperature
    if query.temperature <= 0:
        query.temperature = 0.01
    # tool_choice is set to 'auto' if there is tool

    # set tool_choice if user didn't set it and ensure parameters are correct
    # if no tool, set tool_choice to none
    if query.tools:
        if query.tool_choice is None:
            if query.tools:
                query.tool_choice = "auto"
            else:
                query.tool_choice = "none"
        elif query.tool_choice == "none":
            # wipe out all the tools if any
            query.tools = None
        elif isinstance(query.tool_choice, ToolForce):
            # wipe out all the tool except the tool to be forced 'tool_choice' in api call
            force_single_tool = []
            for tool in query.tools:
                if tool.function.name == query.tool_choice.function.name:
                    force_single_tool.append(tool)

            if len(force_single_tool) != 1:
                raise Exception("tool_choice not exist in function list")
            else:
                query.tool_choice = "required"
                query.tools = force_single_tool
    else:
        query.tool_choice = "none"

    # validate that prefix_strings and regex_prefix_pattern  or regex_pattern can not be used together
    if query.prefix_strings and (query.regex_pattern or query.regex_prefix_pattern):
        raise HTTPException(status_code=400, detail="refix_strings and regex_pattern, regex_prefix_pattern can not be used together")

    return query


async def result_generator(gen_queue):
    completed_event = asyncio.Event()

    while True:
        result = await gen_queue.get()
        if result == {"data": "[DONE]"}:
            completed_event.set()
            yield result
            break

        yield {"data": result.json()}

    # Wait until the task is marked as complete
    await completed_event.wait()


# noinspection PyAsyncCall
@router.post("/v1/chat/completions")
async def chat_completion(request: Request, query: ChatMLQuery):
    # https://platform.openai.com/docs/api-reference/chat/create

    global llm_dict, default_model_name
    gen_queue = GenQueue()      # this queue will hold the result for this generation

    try:
        # validate and fix query
        query = validate_api_request(query)

        if llm_dict.get(query.model):
            model_name_to_use = query.model
        else:
            model_name_to_use = default_model_name

        llm = llm_dict[model_name_to_use]["model"]
        prompt_eng = llm_dict[model_name_to_use]["prompt_engine"]

        # log if thinking is used
        if query.thinking_template:
            logger.info(f"thinking is used with returnThinking set to {query.return_thinking}")

        # start the generation task
        asyncio.create_task(llm.chat(
            query=query,
            prompt_eng=prompt_eng,
            gen_queue=gen_queue,
            request=request,
        ))

        # send the response to client
        if query.stream:
            # EventSourceResponse take iterator so need to handle at here
            if query.artifact == "No":     # not using artefact
                return EventSourceResponse(
                    chat_completion_response_stream(
                        query=query, gen_queue=gen_queue, model_name=model_name_to_use, request=request,
                    ))
            else:
                return EventSourceResponse(
                    chat_completion_response_artifact_stream(
                        query=query, gen_queue=gen_queue, model_name=model_name_to_use, request=request,
                    ))
        else:
            if query.artifact == "No":     # not using artefact
                return await chat_completion_response(query=query, gen_queue=gen_queue, model_name=model_name_to_use, request=request,)
            else:
                return await chat_completion_response_artifact(query=query, gen_queue=gen_queue, model_name=model_name_to_use, request=request,)
    except HTTPException as e:
        logger.error(e)
        return e


@router.post("/v1/chat/generate")
@router.post("/v1/completions")
async def generate(request: Request, query: GenerateQuery):
    global default_model_name
    if query.model and llm_dict.get(query.model):
        model_name_to_use = query.model
    else:
        model_name_to_use = default_model_name

    llm = llm_dict[model_name_to_use]["model"]

    gen_queue = GenQueue()  # this queue will hold the result for this generation

    # start the generation task
    asyncio.create_task(llm.chat_raw(
        prompt=query.prompt,
        max_tokens=query.max_tokens,
        gen_queue=gen_queue,
        request=request,
    ))

    if query.stream:
        # EventSourceResponse take iterator so need to handle at here
        return EventSourceResponse(completion_response_stream(request, gen_queue, model_name=model_name_to_use, request=request,))
    else:
        return await completion_response(gen_queue, model_name=model_name_to_use, request=request,)


@router.post("/v1/embeddings")
async def embeddings(request: Request, query: EmbeddingRequest):
    global default_model_name
    embedding_model = llm_dict[default_model_name]["model"]

    # for embedding, hard enforcement of matching model name
    if query.model != embedding_model.model_name:
        raise HTTPException(status_code=400,
                            detail=f"Embedding model {query.model} is not found. Current loaded model is {embedding_model.model_name}")

    return await embedding_model.text_embeddings(
        query=query,
    )


@router.get("/v1/models")
async def get_models(request: Request):
    data = []
    for model_name in llm_dict.keys():
        data.append(ModelObject(id=model_name))

    return ModelObjectResponse(data=data)


@router.options("/v1/chat/completions")
async def options_handler(request: Request):
    return JSONResponse(
        status_code=200,
        content={"message": "Local OpenAI response for OPTIONS request"},
        headers={
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
            "Access-Control-Allow-Headers": "",
        },
    )


@router.post("/load_model")
def load_model(model_spec: ModelParser):
    global config_manager, llm_dict

    # get the config from the yml
    model_name = model_spec.model_id
    model_config = config_manager.get_model_config(model_name)

    if not model_config:
        raise Exception(f"Model config for '{model_name}' not exist in ~/gallama/model_config.yaml")

    # load the model with config from the model_spec and yml. model_spec comes from cli
    if model_config["backend"] != "embedding":
        prompt_eng = PromptEngine(prompt_format=model_config["prompt_template"])

        if model_spec.draft_model_id:
            draft_model_config = config_manager.get_model_config(model_spec.draft_model_name)
            if not draft_model_config:
                raise Exception(
                    f"Model config for '{model_spec.draft_model_name}' not exist in ~/gallama/model_config.yaml")
        else:
            draft_model_config = {}

        # load LLM model
        llm_base = Model(
            model_spec=model_spec,
            model_config=model_config,
            draft_model_config=draft_model_config,
            eos_token_list_from_prompt_template=prompt_eng.eos_token_list,
        )

        chat_generator_dict = {
            "exllama": ChatGenerator,
            "llama_cpp": ChatGeneratorLlamaCpp,
        }

        chatGenerator_to_use = chat_generator_dict[llm_base.backend]
        llm = chatGenerator_to_use(llm_base)

        # update dict
        llm_dict[model_name] = {
            "model": llm,
            "prompt_engine": prompt_eng,
        }
    else:   # embedding model
        from gallama.backend.embedding import EmbeddingModel

        llm = EmbeddingModel(
            model_id=model_config["model_id"],
            model_name=model_name,
            model_spec=model_spec,
            model_config=model_config,
        )

        # update dict
        llm_dict[model_name] = {
            "model": llm,
        }


    logger.info("Loaded: " + model_name)




@router.post("/delete_model")
def delete_model(model_name: str):

    if model_name in llm_dict.keys():
        del llm_dict[model_name]

        # Clear CUDA cache if using GPU
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        # Run garbage collection
        gc.collect()

        logger.info("Deleted model: " + model_name)


@router.get("/status")
async def get_status():
    logger.info("Status endpoint called")
    return {"status": "ready" if model_ready else "loading"}


# Add a health check endpoint
@router.get("/health")
async def health_check():
    return {"status": "healthy"}


async def startup_event():
    # run some dummy generation so that cache is initialized
    logger.info("Generator initialization")
    global llm_dict
    gen_queue = GenQueue()

    for _model_name, _model in llm_dict.items():
        if _model.get("prompt_engine"):     # this is an LLM not embedding model
            llm = _model["model"]
            await llm.chat_raw(
                prompt=f"{ARTIFACT_SYSTEM_PROMPT}\nWrite a 500 words story on Llama",
                stream=False,
                max_tokens=200,
                gen_queue=gen_queue,
                quiet=True,
                request=None,
            )

    gen_queue = None
    logger.info("Generator warmed up")


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Startup code
    await startup_event()
    yield


def make_server(args):
    global logger
    global model_ready
    global llm_dict, default_model_name
    global draft_spec_dict
    # load yaml file of model info
    logger.info(args)
    model_dict = {}
    draft_spec_dict = {}    # for speculative decoding

    # parse cli argument to get model info
    if args.model_id:
        for item in args.model_id:
            model_dict.update(item)
        model_spec = ModelParser.from_dict(model_dict)

        # set the env for log to zmq
        # Note: This assumes you want to use the first model's ID if multiple are provided
        os.environ['MODEL_NAME'] = model_spec.model_id
        os.environ['MODEL_PORT'] = str(args.port)

    # set logger level
    if args.verbose:
        logger.setLevel(DEBUG)
        os.environ["LOCAL_OPEN_AI_VERBOSE"] = '2'   # turn on verbosity for all

    logger.info("Parsed Arguments:" + str(args))  # Debug statement

    # initialize fastapi

    # set ready status to False
    model_ready = False

    if args.detached:
        # send logging to zmq so that it will show in the parent log
        logger = get_logger(name="child", to_zmq=True, to_console=False)
    else:
        # keep the default logger declared on top
        pass

    #os.environ["PYTORCH_CUDA_ALLOC_CONF"] = 'per_thread_default_stream'

    app = FastAPI(lifespan=lifespan)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"], #origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(router)

    @app.middleware("http")
    @app.middleware("https")
    async def log_requests(request: Request, call_next):
        try:
            if request.method in ("POST", "PUT", "PATCH"):  # Methods that typically have a body
                request_content = await request.body()
                if request_content:  # Only process if the body is not empty
                    request_content = json.dumps(json.loads(request_content.decode("utf-8")), indent=2)
                    logger.debug(f"API Request:\nMethod: {request.method}\nURL: {request.url}\n{request_content}")

            response = await call_next(request)
        except RequestValidationError as e:
            logger.debug(f"Validation error:\n{e}")
            response = JSONResponse(status_code=422, content={"detail": "Validation error"})

        return response

    llm_dict = {}

    # make relevant parameter global for endpoint function to use
    # load LLM model
    if args.model_id:
        model_dict = {}
        for item in args.model_id:
            model_dict.update(item)
        model_spec = ModelParser.from_dict(model_dict)
        default_model_name = model_spec.model_id

        # load model
        load_model(model_spec)

    model_ready = True
    uvicorn.run(app, host=args.host, port=args.port, log_config=None)    # disable log_config to use our custom logger


def parse_dict(arg):
    """Parses a key=value string and returns a dictionary."""
    result = {}
    for pair in arg.split():
        key, value = pair.split('=')
        result[key] = value.strip("'")  # Strip single quotes here as well
    return result


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(description="Launch local OpenAI model")
    arg_parser.add_argument("-id", "--model_id", nargs='+', type=parse_dict, default=None)
    arg_parser.add_argument('-v', "--verbose", action='store_true', help="Turn on more verbose logging")
    arg_parser.add_argument('-d', "--detached", action='store_true', help="Log to ZeroMQ")
    arg_parser.add_argument("--host", type=str, default="127.0.0.1", help="The host to bind to.")
    arg_parser.add_argument('-p', "--port", type=int, default=8000, help="The port to bind to.")
    arg_parser.add_argument("--reload", action="store_true", help="Enable auto-reload.")

    args = arg_parser.parse_args()

    make_server(args)



import time
import sys
import httpx
import uuid
from fastapi import FastAPI, HTTPException, Request, BackgroundTasks
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from collections import defaultdict
from typing import Union
from gallama.data_classes.data_class import ModelParser, ModelObjectResponse, ModelObject, ModelDownloadSpec
from gallama.data_classes import ModelRequest, ModelInstanceInfo,  ModelInfo, AgentWithThinking, StopModelByPort
from gallama.server_engine import download_model_from_hf, handle_mixture_of_agent_request, create_options_response
from typing import List, Dict, Optional
from gallama.config import ConfigManager
from gallama.server_engine import forward_request
import psutil
import shutil
import asyncio
import uvicorn
import argparse
from gallama.logger.logger import get_logger
from gallama.utils.utils import get_package_file_path
from gallama.server_engine import log_model_status
import logging
from logging import DEBUG, INFO
import zmq
import os
import json
import threading
import traceback


manager_app = FastAPI()


@manager_app.middleware("http")
@manager_app.middleware("https")
async def log_requests(request: Request, call_next):
    try:
        if request.method in ("POST", "PUT", "PATCH"):  # Methods that typically have a body
            request_content = await request.body()
            if request_content:  # Only process if the body is not empty
                request_content = json.dumps(json.loads(request_content.decode("utf-8")), indent=2)
                logger.info(f"API Request:\nMethod: {request.method}\nURL: {request.url}\n{request_content}")

        response = await call_next(request)
    except RequestValidationError as e:
        logger.debug(f"Validation error:\n{e}")
        response = JSONResponse(status_code=422, content={"detail": "Validation error"})

    return response

# Add CORS middleware
manager_app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)


models: Dict[str, ModelInfo] = {}
model_load_queue = asyncio.Queue()
loading_lock = asyncio.Lock()
active_requests_lock = asyncio.Lock()
task_status = {}
config_manager = ConfigManager()

# Define a variable for the starting port
START_PORT = 8001

# Increase the timeout for long-running tasks (adjust as needed)
TIMEOUT = 300  # 5 minutes

strict_mode = False

# List of endpoints to exclude from API gateway redirection
EXCLUDED_ENDPOINTS = [
    "/add_model", "/remove_model", "/list_models", "/list_available_models",
    "/v1/add_model", "/v1/remove_model", "/v1/list_models", "/v1/list_available_models",
    "/list_models", "/v1/list_models"
    "/v1/remove_model_by_port","/remove_model_by_port",
    "/v1/models", "/task_status"
]
EMBEDDING_SUBPATHS = []     # No overwriting at the moment
# EMBEDDING_SUBPATHS = [
#     {
#         "original": "/v1/embeddings",
#         "replacement": "/v1/embeddings"
#     },
#     # Add more paths here as needed, e.g.:
#     # {"original": "/v2/some_path", "replacement": "/new_path"}
# ]

# Dictionary to keep track of active requests per instance
active_requests: Dict[int, int] = defaultdict(int)

# Add a semaphore for concurrency control
MAX_CONCURRENT_REQUESTS = 100  # Adjust this value based on your system's capacity
request_semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)


# set up logging
def start_log_receiver(zmq_url):
    context = zmq.Context()
    socket = context.socket(zmq.PULL)
    socket.bind("tcp://*:5555")

    # Initialize the logger for the receiver
    receiver_logger = get_logger(name="log_receiver", to_console=True, to_file=False, to_zmq=False)

    receiver_logger.info(f"Log receiver started on {zmq_url}")

    def receive_logs():
        while True:
            try:
                message = socket.recv_json()
                log_level = getattr(logging, message['level'].upper(), logging.INFO)

                receiver_logger.log(
                    level=log_level,
                    msg=f"{message['model']}:{message['port']} | {message['log']}"
                )
            except zmq.Again:
                # No message available, sleep for a short time
                time.sleep(0.1)
            except zmq.ZMQError as e:
                receiver_logger.error(f"ZMQ Error in log receiver: {e}")
            except json.JSONDecodeError as e:
                receiver_logger.error(f"JSON Decode Error in log receiver: {e}")
            except Exception as e:
                receiver_logger.error(f"Unexpected error in log receiver: {e}")
                receiver_logger.error(traceback.format_exc())

    # Start the receiver in a separate thread
    receiver_thread = threading.Thread(target=receive_logs, daemon=True)
    receiver_thread.start()

    return receiver_thread



# Start the log receiver in a separate thread
DEFAULT_ZMQ_URL = "tcp://127.0.0.1:5555"  # Using 5559 as a standard port for logging


# Initialize the logger for the manager
# Set to_console=True and to_zmq=False to avoid duplication
logger = get_logger(name="manager", to_console=True, to_zmq=False)





@manager_app.post("/download_model")
def download_model(model_spec: ModelDownloadSpec):
    download_model_from_hf(model_spec)


async def model_loader():
    while True:
        logger.info("Model loader waiting for next model in queue")
        model = await model_load_queue.get()
        logger.info(f"Model loader retrieved model from queue: {model}")
        async with loading_lock:
            try:
                logger.info(f"Starting to load model: {model}")
                logger.info(f"Current queue size: {model_load_queue.qsize()}")
                await run_model_with_timeout(model)
            except Exception as e:
                logger.exception(f"Error loading model {model} instance: {str(e)}")
            finally:
                logger.info(f"Finished processing model: {model}")
                model_load_queue.task_done()
                logger.info(f"Remaining queue size: {model_load_queue.qsize()}")

        logger.info("Model loader finished processing, looping back")
        await asyncio.sleep(1)


async def wait_for_model_ready(port, timeout=300):  # 5 minutes timeout
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(f"http://localhost:{port}/health", timeout=5.0)
                if response.status_code == 200:
                    return True
        except httpx.RequestError:
            pass  # The server_engine might not be up yet, so we'll just continue waiting
        await asyncio.sleep(1)
    return False


async def run_model(model: ModelParser):
    try:
        # add the entry for the dictionary (where key is the model name)
        if model.model_name not in models:
            models[model.model_name] = ModelInfo(instances=[])

        # Find the next available port
        port = START_PORT
        while any(instance.port == port for model_info in models.values() for instance in model_info.instances):
            port += 1

        logger.info(f"Attempting to start model {model.model_name} on port {port}")

        model_config = config_manager.configs.get(model.model_name)
        backend = model.backend or model_config.get('backend')

        try:
            # Use the function
            app_path = get_package_file_path('app.py')
            logger.info(f"Using app path: {app_path}")

            # Determine the correct Python executable
            python_exec = shutil.which("python3") or shutil.which("python")

            if backend == "exllama":
                model_cli_args = model.to_arg_string()
                logger.debug(f"model cli: {model_cli_args}")
                process = await asyncio.create_subprocess_exec(
                    python_exec, app_path, "-id", model_cli_args, "--detached", "--port", str(port),
                    stdout=asyncio.subprocess.DEVNULL,
                    # stderr=asyncio.subprocess.DEVNULL,
                )
            else:
                # for embedding infinity doesnt have way to select GPU to be used, hence enforce
                # CUDA_VISIBLE_DEVICES constraints before launching fastapi

                # Create a copy of the current environment
                env = os.environ.copy()

                # Set CUDA_VISIBLE_DEVICES
                env['CUDA_VISIBLE_DEVICES'] = model.get_visible_gpu_indices()

                model_cli_args = model.to_arg_string()
                logger.debug(f"model cli: {model_cli_args}")
                process = await asyncio.create_subprocess_exec(
                    python_exec, app_path, "-id", model_cli_args, "--detached", "--port", str(port),
                    stdout=asyncio.subprocess.DEVNULL,
                    # stderr=asyncio.subprocess.DEVNULL,
                    env=env  # Pass the modified environment to the subprocess
                )

        except Exception as e:
            logger.error(f"Failed to create subprocess for model {model.model_id} on port {port}: {str(e)}")
            return

        # Wait for the model to become ready
        if await wait_for_model_ready(port):
            logger.info(f"Model {model.model_name} on port {port} is ready")
            instance_info = ModelInstanceInfo(
                port=port,
                pid=process.pid,
                status="running",
                model_id=model.model_id,
                embedding=model.backend == "embedding"
            )
            models[model.model_name].instances.append(instance_info)
        else:
            logger.error(f"Timeout waiting for model {model.model_name} on port {port} to become ready")
            await stop_model_instance(model.model_name, port)
            return

        logger.info(f"Model {model.model_id} instance on port {port} is fully loaded and ready")
        log_model_status(models, custom_logger=logger)  # Log status after successfully loading a model

        # Instead of entering an infinite loop, we'll exit the function here
        return

    except Exception as e:
        logger.exception(f"Error running model {model.model_name} instance on port {port}: {str(e)}")
        await stop_model_instance(model.model_name, port)
    finally:
        await cleanup_after_model_load(model.model_name)
        logger.info(f"Exiting run_model for {model.model_name} on port {port}")


async def cleanup_after_model_load(model: str):
    # Perform any necessary cleanup here
    logger.info(f"Performing cleanup after loading model: {model}")
    # For example, you might want to close any open connections or release resources
    # This is a placeholder - add specific cleanup tasks as needed


async def run_model_with_timeout(model: str, timeout: int = 600):  # 10 minutes timeout
    try:
        await asyncio.wait_for(run_model(model), timeout=timeout)
    except asyncio.TimeoutError:
        logger.error(f"Timeout occurred while loading model: {model}")
        # Perform cleanup or error handling here
        await stop_model_instance(model, models[model].instances[-1].port if model in models and models[model].instances else None)


async def stop_model_instance(model: str, port: int):
    if model not in models:
        logger.warning(f"Attempted to stop instance of non-existent model: {model}")
        return

    instance = next((inst for inst in models[model].instances if inst.port == port), None)
    if not instance:
        logger.warning(f"Attempted to stop non-existent instance of model {model} on port {port}")
        return

    try:
        process = psutil.Process(instance.pid)
        process.terminate()
        try:
            process.wait(timeout=10)
        except psutil.TimeoutExpired:
            logger.warning(f"Model {model} instance on port {port} did not terminate gracefully. Forcing kill.")
            process.kill()

        if process.is_running():
            logger.error(f"Failed to kill process for model {model} instance on port {port}")
        else:
            logger.info(f"Successfully stopped model {model} instance on port {port}")
    except psutil.NoSuchProcess:
        logger.warning(f"Process for model {model} instance on port {port} no longer exists")
    except Exception as e:
        logger.exception(f"Error stopping model {model} instance on port {port}: {str(e)}")
    finally:
        models[model].instances = [inst for inst in models[model].instances if inst.port != port]
        if not models[model].instances:
            del models[model]
        logger.info(f"Cleaned up model {model} instance on port {port}")
        log_model_status(models, custom_logger=logger)   # Log status after removing a model instance


@manager_app.post("/v1/stop_model_by_port")
@manager_app.post("/stop_model_by_port")
async def stop_model_by_port(request: StopModelByPort):
    model_to_remove = None
    instance_to_remove = None

    # Find the model and instance by port
    for model, model_data in models.items():
        instance = next((inst for inst in model_data.instances if inst.port == request.port), None)
        if instance:
            model_to_remove = model
            instance_to_remove = instance
            break

    if not model_to_remove or not instance_to_remove:
        logger.warning(f"Attempted to stop non-existent instance on port {request.port}")
        return

    try:
        process = psutil.Process(instance_to_remove.pid)
        process.terminate()
        try:
            process.wait(timeout=10)
        except psutil.TimeoutExpired:
            logger.warning(f"Instance on port {request.port} did not terminate gracefully. Forcing kill.")
            process.kill()

        if process.is_running():
            logger.error(f"Failed to kill process for instance on port {request.port}")
        else:
            logger.info(f"Successfully stopped instance on port {request.port}")
    except psutil.NoSuchProcess:
        logger.warning(f"Process for instance on port {request.port} no longer exists")
    except Exception as e:
        logger.exception(f"Error stopping instance on port {request.port}: {str(e)}")
    finally:
        # Remove the instance from the model
        models[model_to_remove].instances = [
            inst for inst in models[model_to_remove].instances if inst.port != request.port
        ]

        # If no instances remain, remove the model entirely
        if not models[model_to_remove].instances:
            del models[model_to_remove]

        logger.info(f"Cleaned up instance on port {request.port}")
        log_model_status(models, custom_logger=logger)  # Log status after removal


async def get_model_from_body(request: Request) -> str:
    try:
        body = await request.json()
        return body.get("model", "")
    except json.JSONDecodeError:
        return ""


# Add a health check endpoint
@manager_app.get("/health")
async def health_check():
    return {"status": "healthy"}


# Add a periodic health check function
async def periodic_health_check():
    while True:
        await asyncio.sleep(60)  # Check every minute
        for model_name, model_info in models.items():
            for instance in model_info.instances:
                try:
                    async with httpx.AsyncClient() as client:
                        response = await client.get(f"http://localhost:{instance.port}/health", timeout=5.0)
                        if response.status_code != 200:
                            logger.warning(f"Instance {instance.port} of model {model_name} is not healthy")
                            # Implement recovery logic here (e.g., restart the instance)
                except Exception as e:
                    logger.error(f"Error checking health of instance {instance.port} of model {model_name}: {str(e)}")
                    # Implement recovery logic here (e.g., restart the instance)


# Proposed fix:
async def load_models(model_request, task_id):
    task_status[task_id] = "loading"
    try:
        models_to_load = model_request if isinstance(model_request, List) else [model_request]
        for model in models_to_load:
            await model_load_queue.put(model)
            logger.info(f"Model {model} instance queued for loading")

        # Wait for all models to be loaded
        await model_load_queue.join()

        task_status[task_id] = "completed"
    except Exception as e:
        task_status[task_id] = f"failed: {str(e)}"

@manager_app.post("/add_model")
@manager_app.post("/v1/add_model")
async def add_model(model_request: Union[ModelParser, List[ModelParser]], background_tasks: BackgroundTasks, request: Request):
    task_id = str(uuid.uuid4())
    # Log the incoming request body
    logger.info(f"Received raw request: {await request.body()}")


    logger.info(f"Received request to add model(s): {model_request}, Task ID: {task_id}")

    if any(instance.status == "loading" for model in models.values() for instance in model.instances):
        logger.warning("Another model is currently loading. Request rejected.")
        raise HTTPException(status_code=409, detail="Another model is currently loading. Please try again later.")

    task_status[task_id] = "queued"
    background_tasks.add_task(load_models, model_request, task_id)

    return {"message": f"Model(s) queued for loading", "task_id": task_id}


@manager_app.get("/task_status/{task_id}")
@manager_app.get("/v1/task_status/{task_id}")
async def get_task_status(task_id: str):
    status = task_status.get(task_id)
    if not status:
        raise HTTPException(status_code=404, detail="Task not found")
    return {"task_id": task_id, "status": status}


@manager_app.post("/remove_model")
@manager_app.post("/v1/remove_model")
async def remove_model(model_request: ModelRequest):
    if model_request.model_id not in models:
        raise HTTPException(status_code=404, detail="Model not found")
    for instance in models[model_request.model_id].instances:
        await stop_model_instance(model_request.model_id, instance.port)
    return {"message": f"All instances of model {model_request.model_id} removal initiated"}



@manager_app.get("/v1/list_loaded_models")
@manager_app.get("/list_loaded_models")
async def list_models():
    return {name: {"instances": [{"port": inst.port, "status": inst.status} for inst in model.instances]}
            for name, model in models.items()}


@manager_app.get("/list_available_models")
@manager_app.get("/v1/list_available_models")
async def list_available_models():
    config_manager = ConfigManager()
    downloaded_models = config_manager.list_downloaded_models_dict
    # logger.info(downloaded_models)
    return downloaded_models


@manager_app.get("/v1/models")
async def get_models(request: Request):
    data = []
    for name, model in models.items():
        data.append(ModelObject(id=name))

    return ModelObjectResponse(data=data)


@manager_app.get("/loading_status")
async def loading_status():
    return {
        "queue_size": model_load_queue.qsize(),
        "currently_loading": loading_lock.locked(),
        "models": {name: {"instances": [{"port": inst.port, "status": inst.status} for inst in model.instances]}
                   for name, model in models.items()}
    }


async def get_instance_for_model(model: str):
    running_instances = [inst for inst in models[model].instances if inst.status == "running"]
    if not running_instances:
        raise HTTPException(status_code=503, detail=f"No running instances available for model: {model}")

    return min(running_instances, key=lambda inst: active_requests[inst.port])


async def forward_to_multiple_agents(request: Request, agent_list: List[Union[str, AgentWithThinking]], modified_body: str, modified_headers: str):
    tasks = []
    for agent in agent_list:
        if isinstance(agent, str):
            instance = await get_instance_for_model(agent)
            tasks.append(forward_request(request, instance, modified_body, modified_headers))
        elif isinstance(agent, AgentWithThinking):
            instance = await get_instance_for_model(agent.model)
            if agent.thinking_template:
                # Modify the request to include the thinking in the thinking_template field
                agent_body = json.loads(modified_body)
                agent_body["thinking_template"] = agent.thinking_template
                agent_modified_body = json.dumps(agent_body).encode()
                agent_modified_headers = dict(modified_headers)
                agent_modified_headers["content-length"] = str(len(agent_modified_body))
                tasks.append(forward_request(request, instance, agent_modified_body, agent_modified_headers))
            else:
                tasks.append(forward_request(request, instance, modified_body, modified_headers))
    return await asyncio.gather(*tasks)


async def load_balanced_router(request: Request, path: str):
    body = await request.body()
    body_json = json.loads(body)

    if request.url.path in EXCLUDED_ENDPOINTS:
        logger.info(body_json)
        return await request.app.router.get_route_handler()(request)

    if request.method == "OPTIONS":
        return create_options_response(dict(request.headers))

    if body_json.get("mixture_of_agents", False):
        return await handle_mixture_of_agent_request(request, body_json, models, active_requests)

    model = await get_model_from_body(request)

    async with request_semaphore:
        is_embedding = any(subpath["original"] in path for subpath in EMBEDDING_SUBPATHS)

        if not model and strict_mode:
            raise HTTPException(status_code=400, detail="Model must be specified in strict mode")

        available_instances = []

        if strict_mode:
            if model not in models:
                raise HTTPException(status_code=404, detail="Specified model not found")
            available_instances = [inst for inst in models[model].instances if inst.status == "running"]
        else:
            # Try to find a matching model first
            if model:
                if model in models:
                    available_instances = [inst for inst in models[model].instances if inst.status == "running"]

            # If no matching model or no instances found, pick any running instance
            if not available_instances:
                for model_info in models.values():
                    for inst in model_info.instances:
                        if inst.status == "running":
                            if is_embedding:
                                # For embedding requests, select instances with matching model name
                                if not model or inst.model_id == model:
                                    available_instances.append(inst)
                            else:
                                # For non-embedding requests, select all non-embedding instances
                                if not inst.embedding:
                                    available_instances.append(inst)

        if not available_instances:
            raise HTTPException(status_code=503, detail=f"No suitable running instances with requested model '{model}'")

        # Select the instance with the least active requests
        instance = min(available_instances, key=lambda inst: active_requests[inst.port])

        # Increment the active request count for the selected instance
        async with active_requests_lock:
            active_requests[instance.port] += 1

        try:
            logger.info(f"active_requests: {str(active_requests)}")
            logger.info(f"Request routed to model {instance.model_id} instance at port {instance.port}")

            # Forward the request
            response = await forward_request(request, instance)

            return response
        finally:
            # Decrement the active request count after the request is completed
            async with active_requests_lock:
                active_requests[instance.port] -= 1


@manager_app.api_route("/{path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS"])
async def router(request: Request, path: str):
    return await load_balanced_router(request, path)


def shutdown(signal, frame):
    logger.info("Shutting down all models...")
    for model in list(models.keys()):
        for instance in models[model].instances:
            stop_model_instance(model, instance.port)
    logger.info("All models have been shut down. Exiting.")
    sys.exit(0)


async def start_server(port=8000):
    config = uvicorn.Config(manager_app, host="0.0.0.0", port=port, log_level="debug")
    server = uvicorn.Server(config)
    await server.serve()


async def main(model_list=None, port=8000, strict_mode=False):
    receiver_thread = start_log_receiver(DEFAULT_ZMQ_URL)
    logger.info("Starting main function")
    logger.info(f"Strict mode: {'enabled' if strict_mode else 'disabled'}")

    model_loader_task = asyncio.create_task(model_loader())
    logger.info("Created model_loader task")

    # Load initial models
    if model_list:
        logger.info(f"Loading initial models: {[model.model_id for model in model_list]}")
        for model in model_list:
            await model_load_queue.put(model)

        # Wait for all initial models to load
        while not model_load_queue.empty():
            await asyncio.sleep(1)

        logger.info("All initial models have been queued for loading")

    # Start periodic health checks
    health_check_task = asyncio.create_task(periodic_health_check())

    await start_server(port)

    # Ensure tasks are cancelled when the server_engine stops
    model_loader_task.cancel()
    health_check_task.cancel()
    try:
        await asyncio.gather(model_loader_task, health_check_task)
    except asyncio.CancelledError:
        logger.info("Tasks cancelled")


# this parser is different to app.py
def parse_dict(arg):
    """Parses a key=value string and returns a dictionary."""
    result = {}
    for pair in arg.split():
        key, value = pair.split('=')
        result[key] = value.strip("'")  # Strip single quotes here as well
    return result


def llama_picture():
    """ show a fun picture cause why not :) """
    # Load the image
    llama_art = """
        ,,__
        .. \\(oo)
           (__)
             ||----w |
             ||     ||
    """
    logger.info(llama_art)


def run_from_script(args):
    # opening llama picture cause why not
    llama_picture()

    global strict_mode

    if args.strict_mode:
        strict_mode = True

    # parse the cli input
    model_list = []
    if args.model_id:
        for item in args.model_id:
            model_spec = ModelParser.from_dict(item)
            model_list.append(model_spec)
            print(model_spec)

    if model_list:
        logger.info("Initial models: " + str(model_list))

    # set logger level
    if args.verbose:
        logger.setLevel(DEBUG)
    else:
        logger.setLevel(INFO)


    logger.info("Parsed Arguments:" + str(args))

    asyncio.run(
        main(
            model_list=model_list,
            port=args.port,
            strict_mode=args.strict_mode
        )
    )


if __name__ == "__main__":

    logger.info("Script started")
    arg_parser = argparse.ArgumentParser(description="Launch multi model src instance")
    arg_parser.add_argument("--strict_mode", action="store_true", default=False,
                            help="Enable strict mode for routing non-embedding requests to matching model names")
    arg_parser.add_argument("-id", "--model_id", action='append', type=parse_dict, default=None,
                            help="Model configuration. Can be used multiple times for different models. "
                                 "Format: -id model_id=<model_name> [gpu=<vram1>,<vram2>,...] [gpu<n>=<vram>] [cache=<size>] [model_type=exllama|embedding]"
                                 "Examples:\n"
                                 "1. Specify all GPUs: -id model_id=/path/llama3-8b gpu=4,2,0,3 cache=8192\n"
                                 "2. Specify individual GPUs: -id model_id=/path/llama3-8b gpu0=8 gpu1=8 cache=8192\n"
                                 "3. Mix formats: -id model_id=gpt4 gpu=4,2 gpu3=3 cache=8192\n"
                                 "4. Embedding model: -id model_id=/path/gte-large-en-v1.5 gpu=0,0,0,3"
                                 "cache_size is defaulted to model context length if not specified"
                                 "cache_size can be more than context length, which will help model perform better in batched generation"
                                 "cache_size is not application to embedding model"
                                 "VRAM is specified in GB. Cache size is integer which is the context length to cache."
                                 "VRAM for embedding will simple set env parameter to allow infinity_embedding to view the specific GPU and can not enforce VRAM size restriction")
    arg_parser.add_argument('-v', "--verbose", action='store_true', help="Turn on more verbose logging")
    arg_parser.add_argument("--host", type=str, default="127.0.0.1", help="The host to bind to.")
    arg_parser.add_argument('-p', "--port", type=int, default=8000, help="The port to bind to.")

    # arg_parser.add_argument("--reload", action="store_true", help="Enable auto-reload.")

    args = arg_parser.parse_args()

    run_from_script(args)




import uuid
import yaml
import base64
import struct
from lxml import etree
from pathlib import Path

# Lazy import for Exllama
try:
    from exllamav2 import ExLlamaV2Tokenizer
except:
    ExLlamaV2Tokenizer = None

# Lazy import for llama_cpp
llama_cpp = None

def import_llama_cpp():
    global llama_cpp
    if llama_cpp is None:
        try:
            import llama_cpp
        except ImportError:
            raise ImportError("llama_cpp is not installed. Please install it to use Llama models.")
    return llama_cpp

def get_response_uid():
    return "cmpl-" + str(uuid.uuid4().hex)


def get_response_tool_uid():
    return "call_" + str(uuid.uuid4().hex)


def get_token_length(tokenizer, text):
    str_text = text
    if not isinstance(text, str):
        str_text = str(text)

    if ExLlamaV2Tokenizer and isinstance(tokenizer, ExLlamaV2Tokenizer):
        return tokenizer.num_tokens(str_text)
    elif type(tokenizer).__name__ == 'Llama':  # Check the type name instead of using isinstance
        # this is because llama_cpp is optional dependency
        try:
            return len(tokenizer.tokenize(str_text.encode("utf-8"), add_bos=False))
        except AttributeError:
            # If tokenize method is not available, fallback to a general method
            return len(tokenizer.encode(str_text))
    else:
        return len(tokenizer.encode(str_text))

def read_yaml(file_path):
    try:
        with open(file_path, 'r') as file:
            data = yaml.safe_load(file)
            return data
    except FileNotFoundError:
        print(f"The file {file_path} does not exist.")
        return None
    except yaml.YAMLError as exc:
        print(f"Error reading YAML file: {exc}")
        return None


def floats_to_base64(float_array):
    """" convert list of floats to base64 for embedding response """

    # Pack the float array into binary format
    binary_data = struct.pack('f' * len(float_array), *float_array)

    # Encode the binary data to base64
    base64_encoded = base64.b64encode(binary_data)

    # Convert bytes to string for easy handling
    base64_string = base64_encoded.decode('utf-8')

    return base64_string


def parse_xml_to_dict(xml_string):
    # Parse the XML string
    root = etree.fromstring(xml_string)

    # Get the root tag
    root_tag = root.tag

    # Function to convert XML element to a nested dictionary
    def xml_to_dict(element):
        children = list(element)
        if not children:
            return element.text
        result = {}
        for child in children:
            child_dict = xml_to_dict(child)
            if child.tag in result:
                if not isinstance(result[child.tag], list):
                    result[child.tag] = [result[child.tag]]
                result[child.tag].append(child_dict)
            else:
                result[child.tag] = child_dict
        return result

    # Convert the XML to a nested dictionary
    xml_dict = {root.tag: xml_to_dict(root)}

    return xml_dict


def get_package_file_path(file_name: str) -> str:
    """
    Get the absolute path of a file within the gallama package.

    Args:
    file_name (str): The name of the file (e.g., 'app.py')

    Returns:
    str: The absolute path to the file
    """
    current_dir = Path(__file__).parent.parent
    file_path = current_dir / file_name
    if not file_path.exists():
        raise FileNotFoundError(f"File {file_name} not found in the gallama package.")
    return str(file_path.resolve())



ARTIFACT_SYSTEM_PROMPT = """
Artifact System instruction:
The assistant can create and reference artifacts during conversations. Artifacts are for substantial, self-contained content that users might modify or reuse, displayed in a separate UI window for clarity.

Guidelines for structuring your response:

# Good artifacts are:
- Substantial content (>10 lines)
- Content that the user is likely to modify, iterate on, or take ownership of
- Self-contained, complex content that can be understood on its own, without context from the conversation
- Content intended for eventual use outside the conversation (e.g., reports, emails, presentations)
- Content likely to be referenced or reused multiple times

# Don't use artifacts for:
- Simple, informational, or short content, such as brief code snippets, mathematical equations, or small examples
- Primarily explanatory, instructional, or illustrative content, such as examples provided to clarify a concept
- Suggestions, commentary, or feedback on existing artifacts
- Conversational or explanatory content that doesn't represent a standalone piece of work
- Content that is dependent on the current conversational context to be useful
- Content that is unlikely to be modified or iterated upon by the user
- Request from users that appears to be a one-off question

# Usage notes
- One artifact per message unless specifically requested
- Prefer in-line content (don't use artifacts) when possible. Unnecessary use of artifacts can be jarring for users.
- If a user asks the assistant to "draw an SVG" or "make a website," the assistant does not need to explain that it doesn't have these capabilities. Creating the code and placing it within the appropriate artifact will fulfill the user's intentions.
- If asked to generate an image, the assistant can offer an SVG instead. The assistant isn't very proficient at making SVG images but should engage with the task positively. Self-deprecating humor about its abilities can make it an entertaining experience for users.
- The assistant errs on the side of simplicity and avoids overusing artifacts for content that can be effectively presented within the conversation.

artifact_instructions:

Structure your response according to these guidelines:

1. When using artifacts (default for substantial content):
<answer>
    <text>[Simple text here]</text>
    <artifact identifier="[unique-identifier]" type="[artifact_type]" language="[language if applicable]" title="[Brief title]">
    [Your content here]
    </artifact>
    <!-- Additional text or artifact as needed -->
</answer>

2.When explicitly asked NOT to use artifacts or for simple responses:
<answer>
    <text>[Answer here]</text>
</answer>

3. Use the <text> or <artifact> element for each distinct part of your response into chunks where applicable.
4. Use the <text> for content that doesnt form an artifact by the good artifacts criteria mentioned above..
5. For <artifact> elements:
   - The "type" attribute should be one of: "code", or "self_contained_text".
   - Use "code" for code snippets or examples.
   - Use "self_contained_text" for detailed explanations or longer text content, standalone/self-contained elements.
   - Include a unique "identifier" attribute using kebab-case (e.g., "example-code-snippet").
   - Include a "language" attribute for code snippets (e.g., language="python").
   - Include a brief "title" attribute describing the content.
6. Place your content directly within the <artifact> tags.
7. For code snippets, preserve indentation and line breaks.
8. You do not need to escape the content inside the <artifact> tags, as there will be a way to parse and handle that.
9. Both <text> and <artifact> elements of type=self_contained_text will be displayed in a separate UI window with markdown support.
10. If write code inside <text> tag please use ``` to indicate the codeblock and programming language.
    For code inside <artifact>, do not use ``` to indicate.

Example response structure when using artifact (default):
<answer>
    <text>Here is a brief explanation...</text>
    <artifact identifier="example-function" type="code" language="python" title="Example Python Function">
def example_function():
    # code here
    print("Hello, world!")
    </artifact>
    <artifact identifier="detailed-report" type="self_contained_text" title="Comprehensive Topic Analysis">
A detailed report on the topic, which may include multiple paragraphs...
    </artifact>
</answer>

Example response structure when user explicitly ask to not artifact:
<answer>
    <text>Here is a brief explanation...
    ```python
    # code here
    print("Hello, world!")
    ```
    A detailed report on the topic, which may include multiple paragraphs...
    </text>
</answer>

Remember: CDATA is prohibited. The data is processed in a special way and NO XML escaping is needed.
XML comment ( <!-- tag) is prohibited.
<text> and <self_contain_text> will be displayed in a separate UI window using markdown.
Write content directly within the tags.
The XML interpreter only recognize tag above (<answer>, <text>, <artifact>).

Always precede an artifact by acknowledging the user's question or intention using <text>, unless it is unnecessary or explicitly requested other wise.
End of Artifact System instruction.

"""


from .server_dataclass import (
    ModelRequest,
    ModelInstanceInfo,
    ModelInfo,
    AgentWithThinking,
    MixtureOfAgents,
    StopModelByPort
)

from .data_class import (
    ModelDownloadSpec
)



# generator.py
import time
import re
import asyncio
import uuid
import weakref
from typing import List, Union, Literal, Optional
from pydantic import BaseModel, Field
from fastapi import HTTPException, Request
from .model import Model
from gallama.data_classes.data_class import GenerationStats, GenEnd, GenText, GenQueue, ChatMLQuery, GenStart
from .tools import Tools, create_function_models_v2
from dataclasses import dataclass
from gallama.utils.utils import get_token_length
from gallama.logger.logger import logger
from .thinking_template import THINKING_TEMPLATE, Thinking
from gallama.api_response.chat_response import get_response_from_queue
from lmformatenforcer import JsonSchemaParser, RegexParser
from lmformatenforcer.tokenenforcer import TokenEnforcerTokenizerData
from concurrent.futures import ThreadPoolExecutor
from importlib.metadata import version

try:
    from exllamav2 import (
        ExLlamaV2Cache,
        ExLlamaV2Cache_Q4,
    )
    from exllamav2.generator import (
        ExLlamaV2Sampler,
        ExLlamaV2DynamicGeneratorAsync,
        ExLlamaV2DynamicJobAsync,
    )
    from exllamav2.generator.filters import ExLlamaV2PrefixFilter
    from lmformatenforcer.integrations.exllamav2 import (
        ExLlamaV2TokenEnforcerFilter,         # TODO to uncomment this after LMFE take in the changes from turboderp
        build_token_enforcer_tokenizer_data
    )
    from .inference_json_lmfe_wrapper import ExLlamaV2TokenEnforcerFilter as ExLlamaV2TokenEnforcerFilterTemp # TODO to remove this after LMFE take in the changes from turboderp
except:
    ExLlamaV2Cache = None
    ExLlamaV2Cache_Q4 = None
    ExLlamaV2Cache = None
    ExLlamaV2Sampler = None
    ExLlamaV2DynamicGeneratorAsync = None
    ExLlamaV2DynamicJobAsync = None
    ExLlamaV2TokenEnforcerFilter = None
    build_token_enforcer_tokenizer_data = None

try:
    from llama_cpp import LogitsProcessorList
    from lmformatenforcer.integrations.llamacpp import (
        build_llamacpp_logits_processor,
        build_token_enforcer_tokenizer_data as build_token_enforcer_tokenizer_data_llama_cpp
    )
except:
    # llama_cpp optional dependency
    build_llamacpp_logits_processor = None
    build_token_enforcer_tokenizer_data_llama_cpp = None
    LogitsProcessorList = None

assert ExLlamaV2Cache or LogitsProcessorList, "Please install ExllamaV2 or LLama CPP Python as backend"

# experimental support for formatron
try:
    from formatron.formatter import FormatterBuilder
    from formatron.integrations.exllamav2 import create_formatter_filter

except:
    FormatterBuilder = None
    create_formatter_filter = None

TOOL_THINKING = THINKING_TEMPLATE["tool_necessity_evaluation"]


@dataclass
class QueueContext:
    """ helper class for short live handling of multiple Queue"""
    gen_queue: 'weakref.ReferenceType[asyncio.Queue]'
    include_GenStats: bool
    include_GenEnd: bool

    @classmethod
    def create(cls, gen_queue: GenQueue, include_GenStats=True, include_GenEnd=True) -> 'QueueContext':
        return cls(
            gen_queue=weakref.ref(gen_queue),
            include_GenStats=include_GenStats,
            include_GenEnd=include_GenEnd)

    def get_queue(self) -> GenQueue | None:
        return self.gen_queue()


class FormatEnforcer:
    """ this class will help to create filter for generation enforcement"""

    def __init__(self):
        pass

    @staticmethod
    def get_default_engine() -> Literal["formatron", "lm_enforcer"]:
        """ this function will select the format enforcer engine to use if not selected by user"""

        # # use formatron if it is available
        # if FormatterBuilder:
        #     return "formatron"
        # else:
        #     return "formatron"
        #     # return "lm_enforcer"

        return "lm_enforcer"


    def regex(self, regex_pattern: str, filter_engine: Literal[
        "formatron", "lm_enforcer"] = None) -> FormatterBuilder | TokenEnforcerTokenizerData:

        # set the filter engine to use
        if not filter_engine:
            filter_engine = FormatEnforcer.get_default_engine()  # if engine is specified, use it

        # create filter if engine is lm_enforcer
        if filter_engine == "lm_enforcer":
            return RegexParser(regex_pattern)

        # create filter if engine is formatron
        if filter_engine == "formatron":
            f = FormatterBuilder()
            _regex = f.regex(regex_pattern, capture_name='regex')
            f.append_line(f"{_regex}")
            return f

    def json(self, pydantic_model, filter_engine: Literal[
        "formatron", "lm_enforcer"] = None) -> FormatterBuilder | TokenEnforcerTokenizerData:
        """ this function will return the filters for format enforcer to generate json output based on Pyantic model"""

        # set the filter engine to use
        if not filter_engine:
            filter_engine = FormatEnforcer.get_default_engine()  # if engine is specified, use it

        # create filter if engine is lm_enforcer
        # if filter_engine == "lm_enforcer" or filter_engine == "formatron":  # TODO currently formatron and nested pydantic model is having issue
        if filter_engine == "lm_enforcer":  # TODO currently formatron and nested pydantic model is having issue
            json_schema = Tools.replace_refs_with_definitions_v2(pydantic_model.model_json_schema())
            return JsonSchemaParser(json_schema)

        # create filter if engine is formatron
        if filter_engine == "formatron":
            f = FormatterBuilder()
            f.append_line(f"{f.json(pydantic_model, capture_name='json')}")
            return f


class ChatGenerator(Model):
    def __init__(
            self,
            llm_base: Model,
    ):
        # unpack all variables from llm_base
        # refer Model class for details of variable available
        self.__dict__.update(llm_base.__dict__)

        # placeholder
        self.pipeline = None

        # format enforcer
        self.formatter = FormatEnforcer()

    async def chat(self, query: ChatMLQuery, prompt_eng, gen_queue: GenQueue, request: Request):
        chat_method = self.chat_with_tool if query.tools or query.tool_choice != "none" else self.chat_no_tool
        return await chat_method(query=query, prompt_eng=prompt_eng, gen_queue=gen_queue, request=request)

    async def chat_raw(
        self,
        prompt: str,
        gen_queue: asyncio.Queue,
        request: Request,
        stream: bool = False,
        max_tokens: int = None,
        quiet=False,
    ):
        return await self.generate(prompt, max_tokens=max_tokens, gen_queue=gen_queue, quiet=quiet, request=request)

    def validate_token_length(self, token_length):
        # TODO to find max_seq_len for llama cpp from Model
        # if max_seq_len == None meaning there is no token length yet
        if self.max_seq_len and token_length > self.max_seq_len:
            raise HTTPException(status_code=400, detail=f"Token length exceeds max length of {self.max_seq_len}")

    # helper function
    async def check_disconnection(self, request: Request, job,
                                  gen_queue_list: Union[GenQueue, QueueContext, List[QueueContext]]):
        try:
            while True:
                if await request.is_disconnected():
                    logger.info("User disconnected")
                    await job.cancel()

                    # add GenEnd to signal the end of generation
                    chunk = GenEnd()
                    for g_queue in gen_queue_list:
                        try:
                            await g_queue.get_queue().put(chunk)
                        except Exception as e:
                            logger.error(f"Error putting GenEnd into queue: {str(e)}")

                    # break the while loop
                    break

                # Use asyncio.wait_for to implement a timeout
                try:
                    await asyncio.wait_for(asyncio.sleep(1), timeout=1.1)
                except asyncio.TimeoutError:
                    # This allows us to check for cancellation more frequently
                    pass

        except asyncio.CancelledError:
            logger.info("Disconnection check was cancelled")
        except Exception as e:
            logger.error(f"An error occurred in check_disconnection: {str(e)}")
        finally:
            logger.info("Exiting check_disconnection")


    async def chat_no_tool(self, query: ChatMLQuery, prompt_eng, gen_queue, request: Request):

        prompt = prompt_eng.get_prompt(
            query,
            thinking_template=query.thinking_template,
        )

        formatter_prefix_regex = self.formatter.regex(
            query.regex_prefix_pattern) if query.regex_prefix_pattern else None

        formatter_regex = self.formatter.regex(query.regex_pattern) if query.regex_pattern else None

        token_length_prompt = get_token_length(self.tokenizer, prompt)
        self.validate_token_length(token_length_prompt)

        # think template prompting
        if query.thinking_template:
            try:
                thinking = Thinking(query.thinking_template)
            except Exception as e:
                raise HTTPException(status_code=400, detail=f"thinking_template is not valid XML string")

            # Not returning thinking
            thinking_queue = GenQueue()
            if query.return_thinking:
                # return thinking to front end
                queue_group = [
                    QueueContext.create(gen_queue=thinking_queue, include_GenEnd=True, include_GenStats=False),
                    QueueContext.create(gen_queue=gen_queue, include_GenEnd=False, include_GenStats=False)
                ]
            else:
                # not return thinking to front end
                queue_group = [
                    QueueContext.create(gen_queue=thinking_queue, include_GenEnd=True, include_GenStats=False),
                ]

            await self.generate(
                prompt,
                gen_type="thinking",
                gen_queue=queue_group,
                temperature=query.temperature,
                top_p=query.top_p,
                prefix_strings=f"<{thinking.root_tag}>",
                stop_words=thinking.root_key_stop_words,
                request=request,
            )
            thinking_response, _ = await get_response_from_queue(thinking_queue)

            # Get the new prompt with thinking response
            prompt = prompt_eng.get_prompt(
                query,
                thinking_template=query.thinking_template,
                thinking_response=thinking_response,
            )

        # 1st response if there is regex to match the regex pattern
        first_response = ""

        if query.regex_prefix_pattern:
            prefix_queue = GenQueue()
            queue_group = [
                QueueContext.create(gen_queue=prefix_queue, include_GenEnd=True, include_GenStats=False),
                QueueContext.create(gen_queue=gen_queue, include_GenEnd=False, include_GenStats=False)
            ]

            await self.generate(
                prompt,
                gen_queue=queue_group,
                temperature=query.temperature,
                top_p=query.top_p,
                formatter=formatter_prefix_regex,
                prefix_strings=query.prefix_strings,
                request=request,
                # stop_words=query.stop_words,
            )

            first_response, _ = await get_response_from_queue(prefix_queue)

            # append generated content to the full prompt
            prompt = prompt.strip() + first_response.strip()

        # Final generation to return to client
        # set prefix string
        prefix_strings = None if query.regex_prefix_pattern else query.prefix_strings

        # Handle Artifact mode: overwrite prefix_strings if in artifact mode
        stop_words_to_use = query.stop_words
        banned_strings = None
        if query.artifact and query.artifact == "Fast":
            prefix_strings = None
            manual_prefix_string = "<answer>\n "
            prompt += manual_prefix_string

            # ban XML comment format which could mess up the parsing of output
            banned_strings = ["<![CDATA[", "<!--"]

            # add the stopword for artifact tag to the answer
            if isinstance(stop_words_to_use, list):
                stop_words_to_use.append("</answer>")
            elif isinstance(stop_words_to_use, str):
                stop_words_to_use = [stop_words_to_use, "</answer>"]
            else:
                stop_words_to_use = "</answer>"

            # add the initial string as prefix_strings can not be used together with banned_strings
            chunk = GenText(content=manual_prefix_string)
            gen_queue.put_nowait(chunk)

        await self.generate(
            prompt=prompt,
            gen_queue=gen_queue,
            **{
                'temperature': query.temperature,
                'top_p': query.top_p,
                'formatter': formatter_regex,
                'stop_words': stop_words_to_use,
                'max_tokens': query.max_tokens,
                'prefix_strings': prefix_strings,  # already generated as part of the prefix string
                'banned_strings': banned_strings,
                'request': request,
            }
        )

    async def chat_with_tool(self, query: ChatMLQuery, prompt_eng, gen_queue, request: Request):
        # use_tool marker
        use_tool_bool = False  # this will be set to True if tool is used
        fall_back_bool = False  # this will decide if fallback generation with regex enforcement is required

        # tool class have the method to handle converting processing or tool requirement, schema and response
        tool_handler = Tools(
            prompt_eng=prompt_eng,
            tools=query.tools,
            tool_choice=query.tool_choice,
        )

        # substitute the actual list of tool to the thinking template
        tool_thinking_queue = GenQueue()
        tool_thinking_formatted = TOOL_THINKING.xml.format_map(
            {"fstring_available_tools": tool_handler.tool_name_list}
        )

        # perform generation with tool thinking to evaluate if it is necessity to call a tool
        prompt = prompt_eng.get_prompt(
            query,
            pydantic_tool_dict=tool_handler.tool_dict,
            thinking_template=tool_thinking_formatted,
            answer_format_schema=False,
            # leading_prompt=leading_prompt,
        )

        await self.generate(
            prompt,
            gen_queue=tool_thinking_queue,
            temperature=query.temperature,
            top_p=query.top_p,
            stop_words=TOOL_THINKING.root_key_stop_words,
            prefix_strings=f"<{TOOL_THINKING.root_tag}>",
            request=request,
            # formatter=formatter_regex  # no longer enforce format
        )

        # evaluate tool usage necessity
        tool_thinking_response, _ = await get_response_from_queue(tool_thinking_queue)

        # see if llm able to generate the xml format correctly
        try:
            # parse the xml object
            tool_thinking_response_dict = Thinking.parse_xml_to_dict(tool_thinking_response)
            tool_decision = tool_thinking_response_dict[TOOL_THINKING.root_tag]["final_decision"][
                "is_tool_needed"]  # TODO implement a less hardcoding way
            if tool_decision.lower().strip() == "yes":
                use_tool_bool = True
            elif tool_decision.lower().strip() == "no":
                use_tool_bool = False
            else:
                # the format was not enforce, to perform fall back check
                fall_back_bool = True

        except Exception as e:
            logger.error(f"XML parsing failed: {e}")
            # Fallback: check for the presence of Yes/No in the raw XML
            yes_pattern = r'<is_tool_needed>\s*Yes\s*</is_tool_needed>'
            no_pattern = r'<is_tool_needed>\s*No\s*</is_tool_needed>'

            if re.search(yes_pattern, tool_thinking_response, re.IGNORECASE):
                use_tool_bool = True
            elif re.search(no_pattern, tool_thinking_response, re.IGNORECASE):
                use_tool_bool = False
            else:
                fall_back_bool = True

        logger.info(tool_thinking_response)

        # Fall back plan
        fall_back_prompt = ""
        if fall_back_bool:
            logger.info("Tool Analysis fallback")
            # generate fall back response with regex enforcement:
            fall_back_prompt = ("\n Fill in the blank in below sentence with either 'needed' or 'not needed'"
                                "In summary, tool calling is {blank}\n"
                                "Answer: blank=")
            prompt += tool_thinking_response + fall_back_prompt

            # perform generation with tool thinking to evaluate if it is necessity
            tool_thinking_queue_fallback = GenQueue()

            formatter_regex = self.formatter.regex('(needed|not needed)')

            await self.generate(
                prompt,
                gen_queue=tool_thinking_queue_fallback,
                temperature=query.temperature,
                top_p=query.top_p,
                request=request,
                # prefix_strings="n",
                # stop_words=TOOL_THINKING.root_key_stop_words,
                formatter=formatter_regex  # no longer enforce format
            )

            # evaluate tool usage necessity
            tool_thinking_decision_fallback, _ = await get_response_from_queue(tool_thinking_queue_fallback)

            # decide if tool call is required
            if "not needed" in tool_thinking_decision_fallback.lower():
                use_tool_bool = False
            elif "needed" in tool_thinking_decision_fallback.lower():
                use_tool_bool = True

        # USE TOOL
        if use_tool_bool:
            # create the pydantic schema to enforce generation
            tool_combined_pydantic = create_function_models_v2(tool_handler.tool_dict)

            class ToolCalling(BaseModel):
                """ The format to call one or multiple tools """
                functions_calling: List[Union[tuple(tool_combined_pydantic)]] = Field(
                    description='the list of functions to call in chronological order',
                    default=[]
                )

            # class ItemModel(BaseModel):
            #     Use: Literal['Yes', 'No']
            #     reason: str

            # answer_format_schema = tool_handler.replace_refs_with_definitions_v2(ToolCalling.schema())
            #
            # # get format enforcer
            # formatter = JsonSchemaParser(answer_format_schema)

            formatter_json = self.formatter.json(pydantic_model=ToolCalling)

            # Experiment feature, formulate function calling as python programming. Which is more natural than a random Json output as part of conversation
            tool_as_code_prompt = """
def run_function(arg_dict):
    function_calls = arg_dict["functions_calling"]

    if function_calls == []:
        print("No function/tool calling needed")
        return

    for call in function_calls:
        function_name = call["name"]
        arguments = call["arguments"]
        globals()[function_name](**arguments)

# Perform function calling if need to
arg_dict = """

            # get final prompt
            prompt = prompt_eng.get_prompt(
                query,
                thinking_template=TOOL_THINKING.xml,
                thinking_response=tool_thinking_response,
                pydantic_tool_dict=tool_handler.tool_dict,
                answer_format_schema=True,
                leading_prompt=(
                    f"{fall_back_prompt}\n"
                    'Now i will convert my answer above into "functions_calling" format by continuing this continue this code.\n'
                    f"{tool_as_code_prompt}"
                ),
            )
            logger.info(prompt)

            # generate
            await self.generate(
                prompt,
                gen_queue=gen_queue,
                gen_type=GenStart(gen_type="tool"),
                temperature=query.temperature,
                top_p=query.top_p,
                # stop_words=TOOL_THINKING.root_key_stop_words,
                prefix_strings=['{\n "functions_calling": ['],
                formatter=formatter_json,
                max_tokens=query.max_tokens,
                request=request,
            )

        # NOT USE TOOL
        if not use_tool_bool:
            prompt = prompt_eng.get_prompt(query)

            if query.tool_choice == "auto":
                # Normal generation
                await self.generate(
                    prompt,
                    gen_queue=gen_queue,
                    gen_type=GenStart(gen_type="text"),
                    temperature=query.temperature,
                    prefix_strings=query.prefix_strings,
                    max_tokens=query.max_tokens,
                    request=request,
                )
            else:
                # tool choice is forced -> return empty tool calling
                gen_queue.put_nowait(GenStart(gen_type="tool"))
                gen_queue.put_nowait(GenText(content='{"functions_calling":[]}'))
                gen_queue.put_nowait(GenerationStats())
                gen_queue.put_nowait(GenEnd())

    class ExllamaV2Pipeline:
        """ class to hold objects required for Exllama V2 text generation"""

        def __init__(
                self,
                cache: Union[ExLlamaV2Cache, ExLlamaV2Cache_Q4],
                generator: ExLlamaV2DynamicGeneratorAsync,
                lm_enforcer_tokenizer_data: TokenEnforcerTokenizerData,
        ):
            self.cache = cache
            self.generator = generator
            self.lm_enforcer_tokenizer_data = lm_enforcer_tokenizer_data

    async def _get_pipeline_async(self):
        # if not self.cache:
        #     logger.info("Custom Cache size: " + str(self.cache_size))
        #     self.cache = ExLlamaV2Cache_Q4(self.model, max_seq_len=self.cache_size, lazy=not self.model.loaded)

        # Test VRAM allocation with a full-length forward pass
        # input_ids = torch.zeros((1, self.max_seq_len), dtype=torch.long)
        # model.forward(input_ids, cache=cache, preprocess_only=True)

        lm_enforcer_tokenizer_data = build_token_enforcer_tokenizer_data(self.tokenizer)

        generator = ExLlamaV2DynamicGeneratorAsync(
            model=self.model,
            cache=self.cache,
            tokenizer=self.tokenizer,
            max_seq_len=self.max_seq_len,
            draft_model=self.draft_model,
            draft_cache=self.draft_cache,
            num_draft_tokens=5,
        )

        return self.ExllamaV2Pipeline(
            cache=self.cache,
            generator=generator,
            lm_enforcer_tokenizer_data=lm_enforcer_tokenizer_data,
        )

    def _get_exllama_gen_settings(
            self,
            temperature: float = 0.01,
            top_p: float = 0.8,
            **kwargs,
    ):
        # settings
        settings = ExLlamaV2Sampler.Settings()
        settings.temperature = temperature
        settings.min_temp = 0.15
        settings.top_k = 50
        settings.top_p = top_p
        settings.min_p = 0.05
        settings.token_repetition_penalty = 1.1
        settings.token_frequency_penalty = 0.05
        settings.token_repetition_range: int = 1024
        # settings.token_repetition_decay: int = 0.98
        settings.temperature_last = False

        return settings

    @staticmethod
    def get_stop_word(text, stop_words) -> Union[str, None]:
        """ this function will match the stop word used given the text that model ended generation with and a list of stop_words."""

        # sort the list by length to find the longest first
        sorted_stop_words = sorted(stop_words, key=len, reverse=True)

        text = text.lstrip()  # Remove trailing whitespace
        for stop_word in stop_words:
            if stop_word in text:
                return stop_word

        return None

    async def generate(
        self,
        prompt: str,
        gen_queue: Union[GenQueue, QueueContext, List[QueueContext]],
        request: Optional[Request] = None,
        # the generated result will be store to this queue
        gen_type: Union[str, GenStart] = "text",
        temperature: float = 0.01,
        top_p: float = 0.8,
        formatter: FormatterBuilder | TokenEnforcerTokenizerData = None,
        stop_words: Union[List[str], str] = None,
        prefix_strings: Optional[Union[str, List[str]]] = None,
        banned_strings: list[str] | None = None,
        max_tokens: int = None,
        quiet=False,
        **kwargs,
    ) -> (str, GenerationStats):
        try:
            # ensure that generator is initialized
            if self.pipeline is None:
                self.pipeline = await self._get_pipeline_async()

            # make gen_queue to List[QueueContext] for standardize downstream handline
            gen_queue_list = None
            if isinstance(gen_queue, QueueContext):
                gen_queue_list = [gen_queue]
            elif isinstance(gen_queue, GenQueue):
                gen_queue_list = [QueueContext.create(gen_queue, include_GenEnd=True, include_GenStats=True)]
            elif isinstance(gen_queue, list):
                gen_queue_list = gen_queue
                # TODO add validation
                # if any(not isinstance(g_queue, QueueContext) for g_queue in gen_queue_list):
                #     raise Exception("gen_queue must be either a GenQueue, QueueContext or a list of QueueContext")
            else:
                raise Exception("gen_queue must be either a GenQueue, QueueContext or a list of QueueContext")

            if not quiet:
                logger.info("----------------------Prompt---------------\n" + prompt)
                logger.debug("----------------------temperature---------\n" + str(temperature))

            # for async generator, create it as part of the generate job

            # get generation setting
            settings = self._get_exllama_gen_settings(temperature, top_p=top_p)

            # convert prompt to token id
            input_ids = self.tokenizer.encode(prompt)
            self.validate_token_length(len(input_ids[0]))

            # format enforcer
            filters = []
            if formatter:
                if isinstance(formatter, TokenEnforcerTokenizerData) or isinstance(formatter, JsonSchemaParser):  # lm format enforcer
                    exllamva_version = version('exllamav2')
                    if exllamva_version<='0.2.0':
                        filters = [ExLlamaV2TokenEnforcerFilter(
                            character_level_parser=formatter,
                            tokenizer_data=self.pipeline.lm_enforcer_tokenizer_data
                        )]
                    else:
                        filters = [ExLlamaV2TokenEnforcerFilterTemp(
                            model=self.model,
                            tokenizer=self.tokenizer,
                            character_level_parser=formatter,
                        )]
                elif FormatterBuilder and isinstance(formatter, FormatterBuilder):  # formatron
                    filters = [create_formatter_filter(self.model, self.tokenizer, formatter)]
                else:
                    raise "Format enforcer is not correctly initialized"

            if prefix_strings:
                assert isinstance(prefix_strings, str) or (isinstance(prefix_strings, list) and len(prefix_strings) > 0)
                filters.append(ExLlamaV2PrefixFilter(self.model, self.tokenizer, prefix_strings))

            # find stop conditions
            if stop_words:
                if isinstance(stop_words, str):
                    stop_words = [stop_words]

                if not self.eos_token_str:
                    raise Exception("EOS token not set in model_config")
                stop_conditions = self.eos_token_str + stop_words  # concat the 2 list
                logger.debug("stop_words: " + str(stop_conditions))
            else:
                stop_conditions = self.eos_token_str

            job_id = uuid.uuid4().hex

            # logger.info("pending_jobs and active_jobs lists in the ExLlamaV2DynamicGenerator")
            # logger.info(f"job_id: {self.pipeline.generator.jobs}")
            max_tokens_to_use = min(
                self.max_seq_len - len(input_ids[0]),
                max_tokens, 4096) if max_tokens else min(self.max_seq_len - len(input_ids[0]), 4096)

            job = ExLlamaV2DynamicJobAsync(
                generator=self.pipeline.generator,
                input_ids=input_ids,
                max_new_tokens=max_tokens_to_use,
                gen_settings=settings,
                stop_conditions=stop_conditions,  # self.eos_token_id if self.eos_token_id else None,
                banned_strings=banned_strings,
                decode_special_tokens=True,
                filters=filters,
                token_healing=True,
                identifier=job_id,
            )

            # break the pipeline if it takes longer than 1s for 1 iteration
            # async def run_job():

            generate_text = ""
            gen_stats = None
            eos = False

            # kick-start the generation and let down stream know gen type
            if isinstance(gen_type, str):
                gen_type_str = gen_type
                gen_type = GenStart(gen_type=gen_type)
            else:
                gen_type_str = gen_type.gen_type  # get out the generation type in str format

            for g_queue in gen_queue_list:
                g_queue.get_queue().put_nowait(gen_type)

            # Create a task to check for disconnection
            disconnect_check_task = None
            if request:
                disconnect_check_task = asyncio.create_task(self.check_disconnection(request, job, gen_queue_list))

            try :
                # start the generation
                async for result in job:
                    if eos:
                        await job.cancel()
                        break

                    # print(result.get("text", ""))
                    # If we enqueue multiple jobs, an iteration might produce results for any (or all) of them. We could direct
                    # outputs to multiple clients here, using whatever dispatch mechanism, but in this example there will only be
                    # outputs pertaining to the single job started above, and it will all go straight to the console.
                    # assert result["job"] == job

                    # Prefilling/ingesting the prompt may happen over multiple iterations, during which the result will have
                    # a "stage" value of "prefill". We can ignore those results and only use the "streaming" results that will
                    # contain the actual output.
                    # if result["stage"] == "streaming":

                    # Depending on settings, the result dict can contain top-K probabilities, logits and more, but we'll just
                    # grab the output text stream.
                    # generate_text += result.get("text", "")
                    # logger.info(f'{datetime.now()} {result.get("text", "")}')
                    chunk = GenText(content=result.get("text", ""), text_type=gen_type_str)
                    for g_queue in gen_queue_list:
                        g_queue.get_queue().put_nowait(chunk)

                    # logger.info(result.get("text", ""))
                    # logger.info(self.tokenizer.encode(result.get("text", "")))
                    # The "streaming" stage also emits the EOS signal when it occurs. If present, it will accompany a
                    # summary of the job. Print the last packet here to illustrate.
                    if result["eos"]:
                        eos = True

                        # if the stop word occurred is from the stop_words and not model result token -> include in result
                        if stop_words and result.get("held") and result.get("held").get("text"):
                            ending_string = result["held"]["text"].rstrip()

                            if ending_string:
                                # find the stop word that was used to end string
                                stop_word_used = self.get_stop_word(ending_string, stop_words)

                                if stop_word_used:
                                    # end_string is custom token -> return
                                    chunk = GenText(content=stop_word_used, text_type=gen_type_str)
                                    for g_queue in gen_queue_list:
                                        g_queue.get_queue().put_nowait(chunk)
                                else:
                                    # ending token is model eos token
                                    pass

                        gen_stats = GenerationStats(
                            input_tokens_count=result["prompt_tokens"],
                            output_tokens_count=result["new_tokens"],
                            time_to_first_token=result["time_prefill"],
                            time_generate=result["time_generate"],
                        )

                        for g_queue in gen_queue_list:
                            if g_queue.include_GenStats:
                                g_queue.get_queue().put_nowait(gen_stats)

                        # this to signal the end of generation
                        for g_queue in gen_queue_list:
                            if g_queue.include_GenEnd:
                                g_queue.get_queue().put_nowait(GenEnd())
            except Exception as e:
                logger.error(e)
            finally:
                if disconnect_check_task:
                    disconnect_check_task.cancel()
                    try:
                        await disconnect_check_task
                    except asyncio.CancelledError:
                        pass
        except Exception as e:
            logger.error(e)


class ChatGeneratorLlamaCpp(ChatGenerator):
    def __init__(
            self,
            llm_base: Model,
    ):
        # unpack all variables from llm_base
        # refer Model class for details of variable available
        self.__dict__.update(llm_base.__dict__)
        self.pipeline = self._get_pipeline()

    class LLamaCppPipeline:
        """ class to hold objects required for Exllama V2 text generation"""

        def __init__(
                self,
                generator,
                lm_enforcer_tokenizer_data: TokenEnforcerTokenizerData,
        ):
            self.generator = generator
            self.lm_enforcer_tokenizer_data = lm_enforcer_tokenizer_data

    def _get_pipeline(self):
        lm_enforcer_tokenizer_data = build_token_enforcer_tokenizer_data_llama_cpp(self.model)

        return self.LLamaCppPipeline(
            generator=self.model,
            lm_enforcer_tokenizer_data=lm_enforcer_tokenizer_data,
        )

    async def _async_generator(self, sync_generator):
        for item in sync_generator:
            yield item

    def _run_generator_and_queue(self, prompt, logits_processor, max_tokens, temperature, stop, gen_queue_list,
                                 top_p=0.8):
        generator = self.pipeline.generator(
            prompt=prompt,
            logits_processor=logits_processor,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            min_p=0.05,
            stop=stop,
            repeat_penalty=1.1,
            stream=True,
        )

        generate_text = ""
        for chunk in generator:
            chunk_text = GenText(content=chunk['choices'][0]['text'])
            generate_text += chunk_text.content
            for g_queue in gen_queue_list:
                g_queue.get_queue().put_nowait(chunk_text)

        return generate_text

    async def generate(
        self,
        prompt: str,
        gen_queue: Union[GenQueue, QueueContext, List[QueueContext]],
        request: Request,
        # the generated result will be store to this queue
        gen_type: Union[str, GenStart] = GenStart(gen_type="text"),
        temperature: float = 0.01,
        top_p: float = 0.8,
        formatter: FormatterBuilder | TokenEnforcerTokenizerData = None,
        stop_words: Union[List[str], str] = None,
        prefix_strings: Optional[Union[str, List[str]]] = None,
        banned_strings: list[str] | None = None,
        max_tokens: int = None,
        quiet=False,
        **kwargs,
    ):

        if not quiet:
            logger.info("----------------------Prompt---------------\n" + prompt)
            logger.debug("----------------------temperature---------\n" + str(temperature))

        # ensure that generator is initialized
        # if self.pipeline is None:
        #     self.pipeline = await self._get_pipeline_async()

        # make gen_queue to List[QueueContext] for standardize downstream handline
        gen_queue_list = None
        if isinstance(gen_queue, QueueContext):
            gen_queue_list = [gen_queue]
        elif isinstance(gen_queue, GenQueue):
            gen_queue_list = [QueueContext.create(gen_queue, include_GenEnd=True, include_GenStats=True)]
        elif isinstance(gen_queue, list):
            gen_queue_list = gen_queue
            # TODO add validation
            # if any(not isinstance(g_queue, QueueContext) for g_queue in gen_queue_list):
            #     raise Exception("gen_queue must be either a GenQueue, QueueContext or a list of QueueContext")
        else:
            raise Exception("gen_queue must be either a GenQueue, QueueContext or a list of QueueContext")

        # convert prompt to token id
        input_ids = self.tokenizer.tokenize(prompt.encode("utf-8"), add_bos=False)
        self.validate_token_length(len(input_ids))

        # format enforcer
        logits_processors = None
        if formatter:
            logits_processors = LogitsProcessorList([
                build_llamacpp_logits_processor(
                    llm=self.pipeline.lm_enforcer_tokenizer_data,
                    character_level_parser=formatter,
                )
            ])

        start_time = time.time()

        # find stop conditions
        if stop_words:
            if isinstance(stop_words, str):
                stop_words = [stop_words]

            if not self.eos_token_str:
                raise Exception("EOS token not set in model_config")
            stop_conditions = self.eos_token_str + stop_words  # concat the 2 list
            logger.debug("stop_words: " + str(stop_conditions))
        else:
            stop_conditions = self.eos_token_str

        max_tokens_to_use = min(
            self.max_seq_len - len(input_ids),
            max_tokens, 4096) if max_tokens else min(self.max_seq_len - len(input_ids),
                                                     4096
                                                     )

        # kickstart the generation and let down stream know gen type
        if isinstance(gen_type, str):
            gen_type = GenStart(gen_type=gen_type)
        for g_queue in gen_queue_list:
            g_queue.get_queue().put_nowait(gen_type)

        # llama cpp python generator is not async, hence running fake async..
        # Run the synchronous generator in a separate thread
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as pool:
            generate_text = await loop.run_in_executor(
                pool,
                self._run_generator_and_queue,
                prompt, logits_processors, max_tokens_to_use, temperature, stop_conditions, gen_queue_list, top_p
            )

        duration = time.time() - start_time

        gen_stats = GenerationStats()
        for g_queue in gen_queue_list:
            if g_queue.include_GenStats:
                g_queue.get_queue().put_nowait(gen_stats)

        # this to signal the end of generation
        for g_queue in gen_queue_list:
            if g_queue.include_GenEnd:
                g_queue.get_queue().put_nowait(GenEnd())

        logger.debug("----------------------LLM Raw Response---------------\n" + generate_text)



from pydantic import BaseModel, Field, validator, ConfigDict, RootModel, field_validator, constr, model_validator
from typing import Optional, Literal, List, Dict, Union, Any, Type
import asyncio
import uuid
import time
import torch
import re


class TextTag(BaseModel):
    tag_type: Literal["text"] = "text"


class ArtifactTag(BaseModel):
    tag_type: Literal["artifact"] = "artifact"
    artifact_type: Literal["code", "self_contained_text"]
    identifier: str
    title: str
    language: Optional[str] = None


class Query(BaseModel):
    prompt: str
    temperature: Optional[float] = 0.1
    stream: Optional[bool] = False


class ToolCall(BaseModel):
    class FunctionCall(BaseModel):
        arguments: str
        name: str

    id: str
    function: FunctionCall
    type: str = "function"
    index: Optional[int] = None


class BaseMessage(BaseModel):
    role: Literal['system', 'user', 'assistant', 'tool']
    content: Optional[str] = ""
    tool_calls: Optional[List[ToolCall]] = None
    tool_call_id: Optional[str] = None

    @validator('content', pre=True)
    def convert_null_to_empty(cls, value):
        if value is None:
            return ""
        return value


class ParameterProperties(BaseModel):
    type: str = "object"
    description: Optional[str] = None
    enum: Optional[List[str]] = None
    items: Optional[dict] = None   # for array


class ParameterSpec(BaseModel):
    type: str = "object"
    properties: Optional[Dict[str, Any]] = None
    required: List[str] = []


class FunctionSpec(BaseModel):
    name: str
    description: Optional[str] = None
    parameters: ParameterSpec = None


class ToolSpec(BaseModel):
    type: str = "function"
    function: FunctionSpec


class SingleFunctionDict(BaseModel):
    name: str


class ToolForce(BaseModel):
    type: str = "function"
    function: SingleFunctionDict

    class Config:
        extra = "forbid"  # This will prevent extra keys in the dictionary


# test_thinking = """
# <plan>
#   <task>Brief task summary</task>
#   <structure>
#     <c1>[text]: Acknowledge question and introduce answer</c1>
#     <c2>[artifact]: Main content (e.g., code)</c2>
#     <c3>[text]: Explain or elaborate on c2</c3>
#     <c4>[artifact]: Additional content if needed</c4>
#     <c5>[text]: Explain or elaborate on c4</c5>
#     <!-- Add more pairs if needed -->
#   </structure>
# </plan>
# """

class ChatMLQuery(BaseModel):
    class ResponseFormat(BaseModel):
        type: Literal["text", "json_object"]

    class StreamOption(BaseModel):
        include_usage: bool = False

    model: Optional[str] = "Mixtral-8x7B"
    messages: List[BaseMessage]
    temperature: Optional[float] = 0.01
    top_p: float = 0.85
    stream: Optional[bool] = False
    tools: Optional[List[ToolSpec]] = None
    tool_choice: Union[None, Literal["none", "auto", "required"], ToolForce] = None
    tool_call_id: Optional[str] = None
    max_tokens: Optional[int] = None

    # not part of openai api
    leading_prompt: Optional[str] = Field(default="", description="The string to append to the end of the prompt, this will not be part of the generated response")
    prefix_strings: Optional[Union[str, List[str]]] = Field(default=None, description="String or list of strings to start the generation with. Can not be used together with regex_prefix_pattern")
    regex_pattern: Optional[constr(min_length=1)] = None   # regex to enforce
    regex_prefix_pattern: Optional[constr(min_length=1)] = Field(default=None, description="regex to enforce in the beginning of the generation, can not be used together with prefix_string")
    stop_words: Optional[List[str]] = None
    thinking_template: Optional[str] = None
    # thinking_template: Optional[str] = test_thinking
    artifact: Optional[Literal["No", "Fast", "Slow"]] = Field(default="No", description="Normal will parse the streamed output for artifact, whereas Strict is slower and will use format enforcer to enforce")
    #thinking_template: Optional[str] = DEFAULT_THINKING     # the xml template for thinking
    return_thinking: Optional[Literal[False, True, "separate"]] = Field(
        default=False,
        description="Return the generated thinking to front end. False - not return, True - return, 'separate' - return separately as .thinking field. If used together with artifact, True will return as separate."
    )

    # not yet supported options from here # TODO
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict[int, float]] = {}
    top_logprob: int = None
    n: int = 1
    presence_penalty: float = 0
    response_format: Optional[ResponseFormat] = None
    seed: Optional[int] = None
    stream_options: Optional[StreamOption] = None


    @validator('regex_pattern', 'regex_prefix_pattern')
    def validate_regex(cls, v):
        """ this function is to handle special regex patterns """
        if v is not None:
            try:
                re.compile(v)
            except re.error as e:
                raise ValueError(f'Invalid regex pattern: {e}')
        return v



# from here on is answer model for response to api request
class OneTool(BaseModel):
    """The format to use to call one tool"""
    name: str = Field(description='name of the function to use')
    arguments: str


class ToolCallResponse(BaseModel):
    index: Optional[int] = None
    id: Optional[str] = None
    function: OneTool
    type: Optional[str] = "function"


class ToolCalling(BaseModel):
    """ The format to call one or multiple tools """
    functions_calling: List[OneTool] = Field(description='the list of functions to call in chronological order',
                                             default=[])


class ChatResponse(BaseModel):
    role: Literal['system', 'user', 'assistant'] = None
    tool_call_id: Optional[str] = None
    content: Optional[str] = None
    # name: Optional[str] = None
    # function_call: Optional[ToolCalling] = None   # depreciated
    tool_calls: Optional[List[ToolCallResponse]] = None


class ChatMessage(BaseModel):
    role: Literal['system', 'user', 'assistant'] = None
    tool_call_id: Optional[str] = None
    content: Optional[str] = None
    name: Optional[str] = None
    # thinking
    thinking: Optional[str] = None

    # function_call: Optional[ToolCalling] = None   # depreciated
    tool_calls: Optional[List[ToolCallResponse]] = None
    artifact_meta: Union[TextTag, ArtifactTag] = None

    def __str__(self) -> str:
        if self.role == "system":
            return f"system:\n{self.content}\n"

        elif self.role == "function":
            return f"function name={self.name}:\n{self.content}\n"

        elif self.role == "user":
            if self.content is None:
                return "user:\n</s>"
            else:
                return f"user:\n</s>{self.content}\n"

        elif self.role == "assistant":
            if self.content is not None and self.function_call is not None:
                return f"assistant:\n{self.content}\nassistant to={self.function_call.name}:\n{self.function_call.arguments}</s>"

            elif self.function_call is not None:
                return f"assistant to={self.function_call.name}:\n{self.function_call.arguments}</s>"

            elif self.content is None:
                return "assistant"

            else:
                return f"assistant:\n{self.content}\n"
        else:
            raise ValueError(f"Unsupported role: {self.role}")


class UsageResponse(BaseModel):
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0


class Choice(BaseModel):
    index: int = 0
    message: ChatMessage
    logprobs: Union[float, None] = None
    finish_reason: Optional[
        Literal["stop", "length", "tool_calls", "content_filter"]
    ] = "stop"

    @classmethod
    def from_message(cls, message: ChatMessage, finish_reason: str):
        return cls(message=message, finish_reason=finish_reason)


class StreamChoice(BaseModel):
    index: int
    delta: ChatMessage = None
    finish_reason: Optional[Literal["stop", "length", "function_call", "tool_calls"]] = None


class ChatCompletionResponse(BaseModel):
    id: str = Field(description='id of the response', default_factory=lambda: "cmpl-" + str(uuid.uuid4().hex))
    object: Literal["chat.completion", "chat.completion.chunk"] = Field(default="chat.completion")
    created: int = Field(description='timestamp when object was created', default_factory=lambda: int(time.time()))
    model: str = Field(description='name of the model')
    choices: List[Union[Choice, StreamChoice]]
    usage: UsageResponse = None


class CompletionChoice(BaseModel):
    text: str
    index: int
    logprobs: Optional[dict] = None
    finish_reason: Optional[str] = None


class CompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"cmpl-{uuid.uuid4().hex}")
    object: Literal["text_completion"] = "text_completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    system_fingerprint: str = Field(default="fp_44709d6fcb")
    choices: List[CompletionChoice]
    usage: Optional[UsageResponse] = None


class CompletionStreamResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"cmpl-{uuid.uuid4().hex}")
    object: Literal["text_completion"] = "text_completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    system_fingerprint: str = Field(default="fp_44709d6fcb")
    choices: List[CompletionChoice]


# embedding dataclass from here
class EmbeddingRequest(BaseModel):
    """ Request to embedding some text in the input"""
    input: Union[str, List[str], List[List[int]]] = Field(description="text to embed, a str or list of str")
    model: str
    dimension: Optional[int] = None
    encoding_format: Optional[Literal["float", "base64"]] = "float"


class EmbeddingObject(BaseModel):
    index: int = Field(description='index of the embedding', default=0)
    object: str = 'embedding'
    embedding: Union[List[float], str] = Field(description='list of float for embedding vector and str for base64')


class EmbeddingResponse(BaseModel):
    class Usage(BaseModel):
        prompt_tokens: int = Field(description='number of tokens in the prompt')
        total_tokens: int = Field(description='total number of tokens')

    object: str = 'list'
    model: str = Field(description='name of the model')
    usage: Usage
    data: List[EmbeddingObject]


class GenerateQuery(BaseModel):
    prompt: str = Field(description='prompt')
    model: Optional[str] = Field(description='name of the model', default=None)
    stream: Optional[bool] = False
    max_tokens: Optional[int] = Field(description='max number of tokens', default=None)


class GenerationStats(BaseModel):
    input_tokens_count: int = Field(description='input tokens count', default=0)
    output_tokens_count: int = Field(description='output tokens count', default=0)
    time_to_first_token: float = Field(description='time to first token', default=0)
    time_generate: float = Field(description='time to generate tokens', default=0)

    @property
    def generation_speed(self) -> float:
        if self.time_generate > 0:
            return round(self.output_tokens_count / self.time_generate, ndigits=1)
        else:
            return 0

    @property
    def total_time(self) -> float:
        return round(self.time_to_first_token + self.time_generate, ndigits=1)

    @property
    def prefill_speed(self) -> float:
        if self.time_to_first_token > 0:
            return round(self.input_tokens_count / self.time_to_first_token, ndigits=1)
        else:
            return 0

    @property
    def total_tokens_count(self) -> float:
        return self.input_tokens_count + self.output_tokens_count


class ModelObject(BaseModel):
    id: str = Field(description='id of the model')
    object: str = Field(description='object type', default="model")
    owned_by: str = Field(description='object owner', default="remichu")
    created_by: int = Field(description='model creation time', default=1686935002)


class ModelObjectResponse(BaseModel):
    object: str = Field(description='object type', default="list")
    data: List[ModelObject] = []

class GenStart(BaseModel):
    """ this item signal start of generation"""
    model_config = ConfigDict(extra="forbid", validate_assignment=True, protected_namespaces=())  # disable protected_namespaces due to it field use model_ in the name

    gen_type: Literal["text", "tool", "thinking"]  = Field(description='True to signal end of generation', default="text")

class GenEnd(BaseModel):
    """ this item signal end of generation"""
    model_config = ConfigDict(extra="forbid", validate_assignment=True, protected_namespaces=())  # disable protected_namespaces due to it field use model_ in the name

    generation_end: bool = Field(description='True to signal end of generation', default=True)


class GenText(BaseModel):
    model_config = ConfigDict(extra="forbid", validate_assignment=True, protected_namespaces=())  # disable protected_namespaces due to it field use model_ in the name

    text_type: Literal["text", "thinking", "tool"] = "text"
    content: str = Field(description='text or thinking')
    @classmethod
    def __instancecheck__(cls, instance):
        return isinstance(instance, cls)


class GenQueue(asyncio.Queue):
    def __init__(self, maxsize=0, allowed_types: List[str] = [GenText, GenerationStats, GenStart, GenEnd]):
        super().__init__(maxsize)
        self.allowed_types = allowed_types

    def _check_item(self, item):
        if self.allowed_types:
            if not any(isinstance(item, allowed_type) for allowed_type in self.allowed_types):
                raise TypeError(f"Item {item} is not an instance of any allowed types: {self.allowed_types}")
        return item

    def _raise_type_error(self):
        raise TypeError(f"Items must be instances of: {', '.join(self.allowed_types)}")

    async def put(self, item):
        self._check_item(item)
        await super().put(item)

    def put_nowait(self, item):
        self._check_item(item)
        super().put_nowait(item)


class Thinking(BaseModel):
    xml: str = Field(description='xml string')
    regex: str = Field(description='regex string to enforce any value', default=None)



class ModelParser(BaseModel):
    model_id: str = Field(description='id of the model from the yml file')
    model_name: Optional[str] = Field(description='name of the model', default=None)
    gpus: Optional[List[float]] = Field(description='VRam usage for each GPU', default=None)
    cache_size: Optional[int] = Field(default=None, description='The context length for cache text in int. If None, will be set to the model context length')
    cache_quant: Optional[Literal["FP16", "Q4", "Q6", "Q8"]] = Field(default=None, description='the quantization to use for cache, will use Q4 if not specified')
    max_seq_len: Optional[int] = Field(description="max sequence length", default=None)
    backend: Optional[Union[Literal["exllama", "llama_cpp", "embedding"], None]] = Field(description="model engine backend", default=None)
    tensor_parallel: Optional[bool] = Field(description="tensor parallel mode", default=False)

    # speculative decoding
    draft_model_id: Optional[str] = Field(description='id of the draft model', default=None)
    draft_model_name: Optional[str] = Field(description='name of the draft model', default=None)
    draft_gpus: Optional[List[float]] = Field(description='VRam usage for each GPU', default=None)
    draft_cache_size: Optional[int] = Field(description='The context length for cache text in int. If None, will be set to the model context length', default=None)
    draft_cache_quant: Optional[Literal["FP16", "Q4", "Q6", "Q8"]] = Field(default=None, description='the quantization to use for cache, will use Q4 if not specified')
    # backend is assumed to be the same as main model



    # dont allow non recognizable option
    model_config = ConfigDict(extra="forbid", validate_assignment=True, protected_namespaces=())  # disable protected_namespaces due to it field use model_ in the name

    @validator('model_name', pre=True, always=True)
    def set_model_name(cls, v, values):
        if v is None and 'model_id' in values:
            return values['model_id'].split('/')[-1]
        return v

    @validator('gpus', pre=True, always=True)
    def validate_gpus(cls, v):
        if v is None:
            return None
        if isinstance(v, dict):
            # Convert the dict to a list based on GPU IDs
            return [v.get(i, 0.0) for i in range(torch.cuda.device_count())]
        return v

    # TODO this is clasing with embedding cause embedding will set visiable GPU and hence it is not seen anymore in this validator
    # @validator('gpus')
    # def check_gpus(cls, gpus):
    #     if gpus is None:
    #         return None
    #     num_gpus = torch.cuda.device_count()
    #     for gpu_id, vram in enumerate(gpus):
    #         if gpu_id < 0 or gpu_id >= num_gpus:
    #             raise ValueError(f"Invalid GPU ID {gpu_id}. Must be between 0 and {num_gpus - 1}")
    #
    #         if vram < 0:
    #             raise ValueError(f"VRAM usage for GPU {gpu_id} must be a non-negative number")
    #
    #         if vram > 0:
    #             device = torch.cuda.get_device_properties(gpu_id)
    #             total_vram = device.total_memory / (1024 ** 3)  # Convert bytes to GB
    #             if vram > total_vram:
    #                 raise ValueError(
    #                     f"Requested VRAM ({vram} GB) for GPU {gpu_id} exceeds available VRAM ({total_vram:.2f} GB)")

        return gpus

    @classmethod
    def from_dict(cls, input_data: Union[str, Dict[str, Any]]):
        if isinstance(input_data, str):
            # If input is a string, split it into a dictionary
            params = input_data.split()
            input_dict = {}
            for param in params:
                key, value = param.split('=')
                input_dict[key] = value.strip("'")  # Strip single quotes here as well
        else:
            # If input is already a dictionary, use it as is
            input_dict = input_data

        model_id = input_dict.get('model_id')
        if model_id:
            model_id = input_dict.get('model_id').strip("'")  # Remove single quotes if present
        #model_name = input_dict.get('model_name')
        max_seq_len = input_dict.get('max_seq_len', None)
        gpus = input_dict.get('gpus')
        cache_size = input_dict.get('cache_size')
        backend = input_dict.get('backend', None)  # Default to None if not provided
        tensor_parallel = input_dict.get('tp', False)
        if tensor_parallel=="True" or tensor_parallel=="true":
            tensor_parallel = True

        # TODO clean up code and merge config setting into config manager
        if backend == "None":
            backend = None
        cache_quant = input_dict.get('cache_quant', None)

        if gpus:
            gpus = [float(x) for x in gpus.split(',')]

        if cache_size:
            cache_size = int(cache_size)

        # speculative decoding
        draft_model_id = input_dict.get('draft_model_id')
        if draft_model_id:
            draft_model_id = draft_model_id.strip("'")  # Remove single quotes if present
        draft_model_name = input_dict.get('draft_model_name')
        if not draft_model_name and draft_model_id:
            draft_model_name = draft_model_id.split('/')[-1]
        else:
            draft_model_name = None

        draft_gpus = input_dict.get('draft_gpus')
        draft_cache_size = input_dict.get('draft_cache_size')
        draft_cache_quant = input_dict.get('draft_cache_quant', None)

        if draft_gpus:
            draft_gpus = [float(x) for x in draft_gpus.split(',')]

        if draft_cache_size:
            draft_cache_size = int(draft_cache_size)

        # Note: We don't need to set model_name here, as the validator will handle it
        return cls(model_id=model_id, gpus=gpus, cache_size=cache_size, backend=backend, cache_quant=cache_quant,
                   max_seq_len=max_seq_len,
                   tensor_parallel=tensor_parallel,
                   draft_model_id=draft_model_id, draft_model_name=draft_model_name,
                   draft_gpus=draft_gpus, draft_cache_size=draft_cache_size, draft_cache_quant=draft_cache_quant)

    def to_arg_string(self) -> str:
        """
        Generate a command-line argument string based on the instance's attributes.

        Returns:
            str: A string representation of the command-line arguments.
        """
        args = [f"model_id={self.model_id}"]

        if self.model_name != self.model_id.split('/')[-1]:
            args.append(f"model_name={self.model_name}")

        if self.gpus is not None:
            args.append(f"gpus={','.join(str(vram) for vram in self.gpus)}")

        if self.max_seq_len is not None:
            args.append(f"max_seq_len={self.max_seq_len}")

        if self.cache_size is not None:
            args.append(f"cache_size={self.cache_size}")

        if self.cache_quant is not None:
            args.append(f"cache_quant={self.cache_quant}")

        if self.backend != "exllama":  # Only include if it's not the default value
            args.append(f"backend={self.backend}")

        if self.tensor_parallel:
            args.append(f"tp={self.tensor_parallel}")

        # Add draft model parameters
        if self.draft_model_id is not None:
            args.append(f"draft_model_id={self.draft_model_id}")

        if self.draft_model_name is not None and self.draft_model_name != self.draft_model_id.split('/')[-1]:
            args.append(f"draft_model_name={self.draft_model_name}")

        if self.draft_gpus is not None:
            args.append(f"draft_gpus={','.join(str(vram) for vram in self.draft_gpus)}")

        if self.draft_cache_size is not None:
            args.append(f"draft_cache_size={self.draft_cache_size}")

        if self.draft_cache_quant is not None:
            args.append(f"draft_cache_quant={self.draft_cache_quant}")

        return " ".join(args)

    def get_visible_gpu_indices(self) -> str:
        """
        Generate a string of GPU indices based on allocated GPUs.
        If no GPUs are specified, return all available GPU indices.

        Returns:
            str: A comma-separated string of GPU indices with allocated VRAM,
                 or all available GPU indices if none are specified.
        """
        if self.gpus is None:
            import torch
            return ','.join(str(i) for i in range(torch.cuda.device_count()))

        if all(vram == 0 for vram in self.gpus):
            return ""  # No GPUs allocated

        visible_devices = [str(i) for i, vram in enumerate(self.gpus) if vram > 0]
        return ','.join(visible_devices)


class ModelDownloadSpec(BaseModel):
    """ dataclass for model download"""
    model_name: str
    quant: Optional[float] = None
    backend: Literal["exllama", "llama_cpp", "embedding"] = "exllama"

    # disable protected_namespaces due to it field use model_ in the name
    model_config = ConfigDict(extra="forbid", protected_namespaces=())



# import transformers
import torch
from typing import List, Dict
from gallama.logger.logger import logger
from gallama.data_classes.data_class import ModelParser

try:
    from exllamav2 import (
        ExLlamaV2,
        ExLlamaV2Tokenizer,
        ExLlamaV2Cache,
        ExLlamaV2Cache_Q4,
        ExLlamaV2Cache_Q6,
        ExLlamaV2Cache_Q8,
        ExLlamaV2Config
    )
except:
    ExLlamaV2 = None
    ExLlamaV2Tokenizer = None
    ExLlamaV2Cache = None
    ExLlamaV2Cache_Q4 = None
    ExLlamaV2Cache_Q6 = None
    ExLlamaV2Cache_Q8 = None
    ExLlamaV2Config = None

try:
    from llama_cpp import Llama
except:
    # optional dependency
    Llama = None

# experimental feature: tensor parallel
try:
    from exllamav2 import ExLlamaV2Cache_TP
except:
    # optional dependency
    ExLlamaV2Cache_TP = None

assert ExLlamaV2 or Llama, "Please install ExllamaV2 or LLama CPP Python as backend"


class Model:
    """A model class that contain the llm and tokenizer"""
    def __init__(self,
        model_spec:ModelParser,
        model_config: Dict,
        draft_model_config: Dict = {},
        eos_token_list_from_prompt_template: List[str] = []
    ):
        # model_spec capture cli argument
        # model_config is from yml file
        self.model_id = model_config["model_id"]
        self.model_name = model_spec.model_name or model_config["model_name"]
        self.max_seq_len = model_spec.max_seq_len or model_config.get("max_seq_len", None)
        if self.max_seq_len is not None:
            self.max_seq_len = (self.max_seq_len//256) * 256     # for paged attention

        self.gpus = model_spec.gpus or model_config.get("gpus") or "auto"
        self.cache_size = model_spec.cache_size or model_config.get("cache_size") or self.max_seq_len   # default to max_seq_len if not set
        if self.cache_size is not None:
            self.cache_size = (self.cache_size//256) * 256     # for paged attention
            if self.max_seq_len is not None:
                # cache size must be greater or equal to max_seq_len
                self.cache_size = max(self.cache_size, self.max_seq_len)

        self.cache_quant = model_spec.cache_quant or model_config.get("cache_quant") or "Q4"
        self.backend = model_spec.backend or model_config["backend"] or "exllama"
        self.tensor_parallel = model_spec.tensor_parallel or model_config.get("tensor_parallel", False)


        # draft model is via cli only
        self.draft_model_id = draft_model_config.get("model_id")
        self.draft_model_name = model_spec.draft_model_name or None
        self.draft_gpus = model_spec.gpus or draft_model_config.get("gpus") or "auto"
        self.draft_cache_size = self.cache_size   # set to the same as main model
        self.draft_cache_quant = model_spec.draft_cache_quant or draft_model_config.get("cache_quant") or "Q4"
        assert (self.draft_model_id is None) == (self.draft_model_name is None)

        # load model and tokenizer; cache is for exllamav2
        self.model, self.tokenizer, self.cache, self.draft_model, self.draft_cache = self.load_model()

        # TODO, to auto detect
        # get the eos_token_str by merging the default config with anything set by user
        self.eos_token_str = list(set(model_config.get("eos_token_list", []) + eos_token_list_from_prompt_template))
        self.eos_token_ids = self.generate_eos_tokens_id()


    def load_model(self):
        model, tokenizer, cache, draft_model, draft_cache = None, None, None, None, None

        if self.backend=="exllama":
            model, tokenizer, cache = self.load_model_exllama(
                model_id=self.model_id,
                backend=self.backend,
                max_seq_len=self.max_seq_len,
                cache_size=self.cache_size,
                cache_quant=self.cache_quant,
                gpus=self.gpus,
                reserve_vram=self._reserve_vram,
                tensor_parallel=self.tensor_parallel,
            )

            # load model and tokenizer; cache is for exllamav2
            if self.draft_model_id is not None:
                draft_model, _, draft_cache = self.load_model_exllama(
                    model_id=self.draft_model_id,
                    backend=self.backend,
                    max_seq_len=self.max_seq_len,   # draft model max_seq_len must be same as main model
                    cache_size=self.draft_cache_size,
                    cache_quant=self.draft_cache_quant,
                    gpus=self.draft_gpus,
                    reserve_vram=self._reserve_vram,
                )

        elif self.backend=="llama_cpp":
            model, tokenizer, cache = self.load_model_llama_cpp(
                model_id=self.model_id,
                backend=self.backend,
                max_seq_len=self.max_seq_len,
                cache_size=self.cache_size,
                cache_quant=self.cache_quant,
                gpus=self.gpus,
                reserve_vram=self._reserve_vram,
                tensor_parallel=self.tensor_parallel,
                draf_model_id=self.draft_model_id
            )

        return model, tokenizer, cache, draft_model, draft_cache


    def load_model_exllama(self, model_id, backend, cache_size, cache_quant, gpus, reserve_vram, max_seq_len=None, tensor_parallel=False):
        """This function return the model and its tokenizer"""
        logger.info("Loading model: " + model_id)

        cache = None    # in case not a backend with separate cache like llama cpp
        tokenizer = None

        config = ExLlamaV2Config(model_id)
        if max_seq_len is not None:
            config.max_seq_len = max_seq_len
        else:
            # set the self.max_seq_len using model config file as it is None at the moment
            max_seq_len = config.max_seq_len
            self.max_seq_len = config.max_seq_len

        model = ExLlamaV2(config)
        tokenizer = ExLlamaV2Tokenizer(config)

        # a simple dict to help map cache quant
        cache_quant_dict = {
            "FP16": ExLlamaV2Cache,
            "Q4": ExLlamaV2Cache_Q4,
            "Q6": ExLlamaV2Cache_Q6,
            "Q8": ExLlamaV2Cache_Q8,
        }

        # cache size need to minimally max_seq_len size
        cache_size_to_use = cache_size if cache_size else config.max_seq_len
        cache_size_to_use = (cache_size_to_use//256) * 256      # round to multiplier of 256 for paged attention
        # ensure cache_size is minimally max_seq_len
        cache_size_to_use = max(cache_size_to_use, max_seq_len)

        # get the cache quantization to use
        cache_quant_to_use = cache_quant_dict[cache_quant]

        logger.info("max_seq_len: " + str(self.max_seq_len))
        logger.info("cache_size: " + str(cache_size_to_use))
        logger.info("Cache Quantization: " + str(cache_quant))

        assert cache_quant_to_use is not None
        assert (isinstance(gpus, str) and gpus == "auto") or (isinstance(gpus, list)), \
            "Device map should be either 'auto', 'gpu' split"

        if not tensor_parallel:
            if isinstance(gpus, str) and gpus == "auto":
                cache = cache_quant_to_use(model, max_seq_len=cache_size_to_use, lazy=True)
                model.load_autosplit(cache, reserve_vram=reserve_vram, progress=True)
            elif isinstance(gpus, list):      # user specify the gpus split
                logger.info("Custom GPU Allocation in GB: " + str(gpus))
                model.load(gpu_split=gpus, progress=True)
                cache = cache_quant_to_use(model, max_seq_len=cache_size_to_use, lazy=not model.loaded)
        else:
            # tensor parallel mode
            logger.info("ExllamaV2 Tensor Parallel enabled")
            if ExLlamaV2Cache_TP:       # ensure that tensor parallel is available
                model.load_tp(progress=True, gpu_split = gpus if isinstance(gpus, list) else None)
                cache = ExLlamaV2Cache_TP(
                    model,
                    max_seq_len = cache_size_to_use,
                    base = cache_quant_to_use,
                )
            else:
                raise ValueError("ExllamaV2 was not installed with tensor parallel")


        return model, tokenizer, cache


    def load_model_llama_cpp(self, model_id, backend, cache_size, cache_quant, gpus, reserve_vram, max_seq_len=None,
                           tensor_parallel=False, draf_model_id=None):
        """This function return the model and its tokenizer"""
        logger.info("Loading model: " + model_id)

        cache = None  # in case not a backend with separate cache like llama cpp
        tokenizer = None

        # TODO to add equivalent support for all exllama option
        # currently speculative decoding not supported by model specific for LLama CPP python
        if isinstance(gpus, str) and gpus == "auto":
            model = Llama(
                model_path=self.model_id,
                n_gpu_layers=-1,
                seed=1,
                n_ctx=self.max_seq_len if self.max_seq_len else 0,  # Uncomment to increase the context window
                flash_attn=True,
                offload_kqv=True,
                # draft_model=draf_model_id,
            )
        elif isinstance(gpus, list):  # user specify the gpus split
            model = Llama(
                model_path=self.model_id,
                n_gpu_layers=-1,
                seed=1,
                n_ctx=self.max_seq_len if self.max_seq_len else 0,  # Uncomment to increase the context window
                flash_attn=True,
                offload_kqv=True,
                tensor_split=gpus,
                # draf_model_id=draf_model_id,
            )
        else:
            raise ValueError("Device map should be either 'auto', 'gpu' split")

        # set max_seq_len based on model
        self.max_seq_len = model._model.n_ctx_train()
        tokenizer = model

        return model, tokenizer, cache


    @property
    def _reserve_vram(self):
        try:
            reserve_block_size = 1024 ** 2
            num_devices = torch.cuda.device_count()
            #reserved_vram = [192 * 1024**2] + [64 * 1024**2] * (num_devices - 1)
            #reserved_vram = [256 * 1024 ** 2] + [96 * 1024 ** 2] * (num_devices - 1)

            # GPU1 is the main GPU for my PC
            # The below is lower threshold than exllamav2 default setting
            reserve_per_gpu = [32 for _ in range(num_devices)]
            main_gpu = 0    # TODO pass it to front end
            reserve_per_gpu[main_gpu] = 64
            reserved_vram = [_reserve * reserve_block_size for _reserve in reserve_per_gpu]
            return reserved_vram
        except:
            # for non cuda env e.g. macbook
            return None


    def generate_eos_tokens_id(self):
        if self.eos_token_str:
            # exllama
            if ExLlamaV2Tokenizer and isinstance(self.tokenizer, ExLlamaV2Tokenizer):
                eos_token_ids = [self.tokenizer.single_id(token) for token in self.eos_token_str]
                return eos_token_ids
            elif self.backend in ["llama_cpp"]:
                return []   # no tokenizer for llama cpp
            else:
                # transformer:
                eos_token_ids = [self.tokenizer.convert_tokens_to_ids(token) for token in self.eos_token_str]
                return eos_token_ids
        else:
            return []



from gallama.logger import logger
import json
from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import Response, StreamingResponse
from gallama.data_classes import ModelInstanceInfo
from typing import Union
import httpx


def create_options_response(headers: dict) -> Response:
    options_headers = {
        "Access-Control-Allow-Origin": headers.get("Origin", "*"),
        "Access-Control-Allow-Methods": "GET, POST, PUT, DELETE, OPTIONS",
        "Access-Control-Allow-Headers": "Content-Type, Authorization, X-Requested-With",
        "Access-Control-Max-Age": "3600",
    }
    return Response(content="", status_code=204, headers=options_headers)


async def stream_response(method: str, url: str, headers: dict, body: bytes):   # TODO currently error while streaming was not returned to client
    try:
        async with httpx.AsyncClient() as client:
            async with client.stream(method, url, headers=headers, content=body, timeout=None) as response:
                if response.status_code >= 400:
                    error_message = json.dumps({"error": f"Server responded with status code: {response.status_code}"})
                    yield error_message.encode('utf-8')
                    return

                async for chunk in response.aiter_bytes():
                    yield chunk
    except httpx.RequestError as exc:
        error_message = json.dumps({"error": f"An error occurred: {str(exc)}"})
        yield error_message.encode('utf-8')
    except Exception as exc:
        error_message = json.dumps({"error": f"An unexpected error occurred: {str(exc)}"})
        yield error_message.encode('utf-8')


async def forward_request(request: Request, instance: ModelInstanceInfo, modified_body: str = None, modified_headers: str = None) -> Union[Response, StreamingResponse]:
    original_path = request.url.path
    path = original_path

    # # Check if the path matches any in EMBEDDING_SUBPATHS and replace if necessary
    # for subpath in EMBEDDING_SUBPATHS:
    #     if path.startswith(subpath["original"]):
    #         path = path.replace(subpath["original"], subpath["replacement"], 1)
    #         logger.info(f"Path replaced: {original_path} -> {path}")
    #         break

    url = f"http://localhost:{instance.port}{path}"
    logger.debug(f"Forwarding request to URL: {url}")

    headers = modified_headers if modified_headers is not None else dict(request.headers)
    body = modified_body if modified_body is not None else await request.body()
    body_json = json.loads(body)

    request.state.instance_port = instance.port

    # Log relevant information
    logger.debug(f"Request method: {request.method}")
    logger.debug(f"Accept header: {headers.get('accept')}")
    logger.debug(f"Content-Type header: {headers.get('content-type')}")
    logger.debug(f"Request body: {body_json}")

    # Check if it's a streaming request
    is_streaming_request = body_json.get('stream', False)
    logger.debug(f"Is streaming request: {is_streaming_request}")

    if request.method == "OPTIONS":
        return create_options_response(headers)
    elif is_streaming_request:
        logger.info("Handling as streaming request")
        return StreamingResponse(
            stream_response(request.method, url, headers, body),
            media_type="application/json"
        )
    else:
        logger.info("Handling as non-streaming request")
        async with httpx.AsyncClient() as client:
            try:
                response = await client.request(method=request.method, url=url, headers=headers, content=body, timeout=None)
                logger.info(f"Response status code: {response.status_code}")
                return Response(content=response.content, status_code=response.status_code,
                                headers=dict(response.headers))
            except httpx.RequestError as exc:
                logger.error(f"An error occurred while forwarding the request to instance at port {instance.port}: {exc}")
                raise HTTPException(status_code=500, detail="Internal server_engine error")






import json
import yaml
from typing import List, Dict
from gallama.data_classes.data_class import BaseMessage, ChatMLQuery, ToolCall
from pydantic import BaseModel
from copy import deepcopy
from gallama.utils.utils import parse_xml_to_dict
from fastapi import HTTPException
from pathlib import Path
from textwrap import dedent
from ..data import ARTIFACT_SYSTEM_PROMPT


class PromptEngine:
    def __init__(self, prompt_format: str):
        self.system_msg_enabled = False
        self.tool_enabled = False
        self.model_prompt_all = self.get_prompt_token()
        if not self.model_prompt_all.get(prompt_format):
            raise ValueError(f'Prompt format {prompt_format} not found in data/model_token.yaml')

        self.model_prompt = self.model_prompt_all.get(prompt_format)
        self.system_msg_enabled = self.model_prompt.get("system_msg_enabled")
        self.tool_enabled = self.model_prompt.get("tool_enabled")
        self.eos_token_list = self.model_prompt.get("eos_token_list")

    @staticmethod
    def get_prompt_token() -> Dict:
        """Get the absolute path to the data directory."""
        yaml_file = Path(__file__).parent.parent / 'data' / 'model_token.yaml'
        with open(yaml_file, 'r') as file:
            yaml_data = yaml.safe_load(file)
        return yaml_data

    def get_conversation_start_token(self):
        return self.model_prompt.get("conversation_start", "")

    def get_conversation_end_token(self):
        return self.model_prompt.get("conversation_end", "")

    def leading_prompt_token(self):
        return self.model_prompt.get("leading_prompt_token", "")

    def get_user_start_token(self):
        return self.model_prompt.get("user_start", "")

    def get_user_end_token(self):
        return self.model_prompt.get("user_end", "")

    def get_sys_start_token(self):
        return self.model_prompt.get("system_start", "")

    def get_sys_end_token(self):
        return self.model_prompt.get("system_end", "")

    def get_assistant_start_token(self):
        return self.model_prompt.get("assistant_start", "")

    def get_assistant_end_token(self):
        return self.model_prompt.get("assistant_end", "")

    def get_tool_start_token(self):
        return self.model_prompt.get("tool_start", "")

    def get_tool_end_token(self):
        return self.model_prompt.get("tool_end", "")

    def get_tool_result_start_token(self):
        return self.model_prompt.get("tool_result_start", "")

    def get_tool_result_end_token(self):
        return self.model_prompt.get("tool_result_end", "")

    def get_tool_call_start_token(self):
        return self.model_prompt.get("tool_call_start", "")

    def get_tool_call_end_token(self):
        return self.model_prompt.get("tool_call_start", "")

    def _get_role_token(self, role, token_type: str):
        if token_type == "start":
            if role == "system":
                return self.get_sys_start_token()
            elif role == "user":
                return self.get_user_start_token()
            elif role == "assistant":
                return self.get_assistant_start_token()
            elif role == "tool":
                return self.get_tool_start_token()
            elif role == "tool_result":
                return self.get_tool_result_start_token()
            elif role == "tool_call":
                return self.get_tool_call_start_token()
        elif token_type == "end":
            if role == "system":
                return self.get_sys_end_token()
            elif role == "user":
                return self.get_user_end_token()
            elif role == "assistant":
                return self.get_assistant_end_token()
            elif role == "tool":
                return self.get_tool_end_token()
            elif role == "tool_result":
                return self.get_tool_result_end_token()
            elif role == "tool_call":
                return self.get_tool_call_end_token()
        else:
            return ""

    @staticmethod
    def _get_message_type(msg: BaseMessage) -> str:
        if msg.tool_call_id:
            return "tool_result"
        elif msg.role == "tool":
            return "tool_call"
        elif msg.role == "assistant":
            return "assistant"
        elif msg.role == "system":
            return "system"
        elif msg.role == "user":
            return "user"
        else:
            raise ValueError(f"Unknown message type {msg.role}")

    def _format_tool_call(self, tool_call: ToolCall) -> str:
        """
            get string representation of tool call like normal python function calling
            Example Output: get_current_weather(location='Boston', unit='Fahrenheit')
        """

        func_name = tool_call.function.name
        args_str = tool_call.function.arguments.replace('"', '')
        return f"{func_name}({args_str})"

    def _format_tool_result(self, msg: BaseMessage) -> str:
        content = msg.content if msg.content else ""

        try:
            content = f"Result of tool call reference id {msg.tool_call_id}:\n" + str(json.dumps(json.loads(content), indent=2)) + "\n\n"
        except:
            content = f"Result of tool call reference id {msg.tool_call_id}:\n" + str(content) + "\n\n"

        return content

    def _format_tool_call_msg(self, msg: BaseMessage) -> str:
        """one msg might have multiple tool calls"""
        tool_call_str = f"Please help to call these tool:\n"

        for tool_call in msg.tool_calls:
            tool_call_str += f"Request for tool call with reference id {tool_call.id}:\n"
            tool_call_str += self._format_tool_call(tool_call)
            tool_call_str += "\n\n"

        return tool_call_str

    def _format_tool_msg(self, pydantic_tool_list: Dict[str, BaseModel]) -> str:
        # append tools calling if it is a prompt
        tools_json = ("\nBelow are the functions available to you to use.\n"
                      "If you need to use multiple functions, please provide it in chronological order.\n"
                      "If user or system prompt request certain restriction (examples: can only use one tool, "
                      "must use a specific tool etc; then please respect the instruction.)\n\n")

        if self.tool_enabled:
            tools_json += self.get_tool_start_token()

        for tool_name, tool in pydantic_tool_list.items():
            tools_json = tools_json + tool_name + ":\n" + str(json.dumps(tool.schema(), indent=2)) + "\n---\n"

        if self.tool_enabled:
            tools_json += self.get_tool_end_token()

        return tools_json

    def _get_one_msg(self, msg: BaseMessage) -> str:
        content = msg.content if msg.content else ""

        prompt = ""
        tool_call_str = ""

        if self._get_message_type(msg) == "tool_result":
            return self._format_tool_result(msg)
        elif self._get_message_type(msg) == "tool_call":
            return self._format_tool_call_msg(msg)
        elif self._get_message_type(msg) in ["assistant", "system", "user"]:
            return content

    def _get_one_msg_grp(self, grp: List[BaseMessage]) -> str:
        prompt = self._get_role_token(role=self._get_message_type(grp[0]), token_type="start")

        for msg in grp:
            prompt = prompt + self._get_one_msg(msg) + "\n\n"

        prompt += self._get_role_token(role=self._get_message_type(grp[0]), token_type="end")
        return prompt

    def _regroup_msg(self, msg_List: List[BaseMessage]):
        # group msg with same role into one array
        regrouped_msg = []
        temp_array = []

        # first create a list of msg with the role converted
        for idx, msg in enumerate(msg_List):
            temp_msg = deepcopy(msg)
            if temp_msg.role == "tool" and not self.tool_enabled:
                temp_msg.role = "user"      # convert to user as this LLM model does not have tool role
            elif temp_msg.role == "system" and not self.system_msg_enabled:
                temp_msg.role = "user"  # convert to user as this LLM model does not have system role

            temp_array.append(temp_msg)

        # now grouping message that was from the same role together
        # this is cause model might have been trained with alternated role prompting
        # and repeat of the same role will not give good response
        previous_msg = msg_List[0]
        grouped_array = []
        for idx, msg in enumerate(temp_array):
            if msg.role == previous_msg.role:
                grouped_array.append(msg)
            else:
                if grouped_array:
                    # create new array
                    regrouped_msg.append(grouped_array.copy())
                    grouped_array = [msg]     # reset temp array
                else:
                    grouped_array.append(msg)

            previous_msg = msg

        # add the last group
        if temp_array:
            regrouped_msg.append(grouped_array.copy())

        # logger.debug("overall regroup:\n" + str(regrouped_msg))
        return regrouped_msg

    def get_prompt(
        self,
        query: ChatMLQuery,
        pydantic_tool_dict: List[BaseModel] = None,
        answer_format_schema: bool = True,  # whether to add the instruction for tool calling answer schema
        leading_prompt: str = None,
        use_thinking: bool = True,
        thinking_template: str = None,
        thinking_response: str = None,
        #prompt_mode: Literal["auto"]
    ) -> str:

        def _create_leading_prompt(original_prompt, query_leading_prompt, model_leading_prompt):
            # add leading prompt from user or model
            # this is usually the role e.g. assistant:, Tom to Jerry:
            _leading_prompt = ""
            if query_leading_prompt:
                _leading_prompt = model_leading_prompt + query_leading_prompt
            elif model_leading_prompt:
                _leading_prompt = model_leading_prompt

            if not _leading_prompt.endswith("\n"):
                _leading_prompt += "\n"

            return _leading_prompt

        thinking_example = ""

        # thinking_example = dedent("""
        # ### Thinking example:
        # Question: Is dog or cat faster? Answer in Capital letter
        # Apply thinking template:
        # <format_restriction>Any specific format requirement from user</format_restriction>
        # My thought process using the thinking template:
        # <format_restriction>User request for answer in capital letter</format_restriction>
        # Final answer:
        # CAT
        # End of Thinking Example.
        # """).strip()


        prompt = self.get_conversation_start_token()     # use arrange to story prompt

        # add the default system prompt for artifact
        if query.artifact != "No":
            prompt += ARTIFACT_SYSTEM_PROMPT

        msg_groups = self._regroup_msg(query.messages)

        # concat all msg
        for idx, msg_group in enumerate(msg_groups):
            prompt = prompt + self._get_one_msg_grp(msg_group)

        # if there is tool, return the prompt with tool
        if pydantic_tool_dict and query.tool_choice != "none":
            prompt = prompt + self._format_tool_msg(pydantic_tool_dict)

            if answer_format_schema:
                prompt += """
IMPORTANT: If you use tool/ functions, please answer using the following schema using json format.
Each item in the array under "functions_calling" is the "name" of the function to call and its "arguments".

{
  "functions_calling": [
    {
      "name": the name of function/ tool to call,
      "arguments": {argument_name: argument_value}
    }
}


If no functions_calling/ tool needed, the response can be:
{
  "functions_calling": []
}
End of Example of answer with Tool/ Function_calling usage.

"""
        # TODO move the thinking_template check to request receiver
        # initialize thinking_template_dict
        thinking_template_dict = {}
        root_key = None
        if thinking_template:
            try:
                thinking_template_dict = parse_xml_to_dict(thinking_template)
                if len(list(thinking_template_dict.keys())) > 1:
                    raise HTTPException(status_code=400, detail=f"thinking_template must be a XML with only 1 root key")

            except Exception as e:
                raise HTTPException(status_code=400, detail=f"thinking_template is not valid XML string")

            # find the XML root key
            root_key = list(thinking_template_dict.keys())[0]

        # prompt creation with thinking_template
        if use_thinking and thinking_template and not thinking_response:
            prompt += _create_leading_prompt(prompt, query.leading_prompt, self.leading_prompt_token())

            prompt += "\nNow, before answering the question, I am required to apply XML thinking template to guide my internal thinking process.\n" + \
                      f"{thinking_example}\n" + \
                      "The thinking template i need to apply to answer this question is as follow:\n" + \
                      thinking_template + "\n" + \
                      f"My thinking using the above XML template as follow:\n\n```xml"
            # add ending token
            # prompt += self.get_conversation_end_token()

        elif use_thinking and thinking_template and thinking_response:
            # the thinking result is ready here
            # root key need to be added as we use it for leading prompt in the prompt creation above so it is not part of the response
            prompt += "\nNow, before answering the question, I am required to apply XML thinking template to guide my internal thinking process.\n" + \
                      f"{thinking_example}\n" + \
                      "Now, the thinking template i need to apply to answer this question is:\n" + \
                      thinking_template + "\n" + \
                      f"My thinking using the XML template as follow:\n```xml\n{thinking_response}\n" + \
                      "Now answer the question. Remember that the thinking above is INVISIBLE to user, " + \
                      "and you are PROHIBITED to mentioned to user about the existence of Thinking section. You are ALLOWED to reiterate points from thinking section to user if necessary."

            # add ending token
            prompt += self.get_conversation_end_token()

            # for thinking response, we put it before the leading prompt
            prompt += _create_leading_prompt(prompt, query.leading_prompt, self.leading_prompt_token())

        else:
            # add ending token
            prompt += self.get_conversation_end_token()

            prompt += _create_leading_prompt(prompt, query.leading_prompt, self.leading_prompt_token())


        # add leading_prompt from code which is usually tool related
        # e.g. ```json
        if leading_prompt:      # leading prompt that passed to this function take highest priority
            prompt += leading_prompt + "\n"

        return prompt

        # match tool call result #TODO

